"""
Functions to generate the queries used in components that are neither graphs or
grid.
"""

QUALSTAT_FILTER_RATIO = """CASE
            WHEN sum(execution_count) = 0 THEN 0
            ELSE sum(nbfiltered) / sum(execution_count)::numeric * 100
        END"""


def qualstat_base_statdata(eval_type=None):
    if eval_type is not None:
        base_cols = ["srvid", "qualid", "queryid", "dbid", "userid"]

        pqnh = """(
        SELECT {outer_cols}
            FROM (
                SELECT {inner_cols}
                FROM {{powa}}.powa_qualstats_quals
            ) expanded
            WHERE (qual).eval_type = '{eval_type}'
            GROUP BY {base_cols}
        )""".format(
            outer_cols=", ".join(base_cols + ["array_agg(qual) AS quals"]),
            inner_cols=", ".join(base_cols + ["unnest(quals) AS qual"]),
            base_cols=", ".join(base_cols),
            eval_type=eval_type,
        )
    else:
        pqnh = "{powa}.powa_qualstats_quals"

    base_query = """
    (
    SELECT srvid, qualid, queryid, dbid, userid, (unnested.records).*
    FROM (
        SELECT pqnh.srvid, pqnh.qualid, pqnh.queryid, pqnh.dbid, pqnh.userid,
          pqnh.coalesce_range, unnest(records) AS records
        FROM {{powa}}.powa_qualstats_quals_history pqnh
        WHERE coalesce_range  && tstzrange(%(from)s, %(to)s, '[]')
        AND pqnh.srvid = %(server)s
    ) AS unnested
    WHERE  (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')
    UNION ALL
    SELECT pqnc.srvid, qualid, queryid, dbid, userid, pqnc.ts, pqnc.occurences,
      pqnc.execution_count, pqnc.nbfiltered,
      pqnc.mean_err_estimate_ratio, pqnc.mean_err_estimate_num
    FROM {{powa}}.powa_qualstats_quals_history_current pqnc
    WHERE pqnc.ts <@ tstzrange(%(from)s, %(to)s, '[]')
    AND pqnc.srvid = %(server)s
    ) h
    JOIN {pqnh} AS pqnh USING (srvid, queryid, qualid)""".format(pqnh=pqnh)

    return base_query


def qualstat_getstatdata(
    eval_type=None,
    extra_from="",
    extra_join="",
    extra_select=[],
    extra_where=[],
    extra_groupby=[],
    extra_having=[],
):
    base_query = qualstat_base_statdata(eval_type)

    # Reformat extra_select, extra_where, extra_groupby and extra_having to be
    # plain additional SQL clauses.
    if len(extra_select) > 0:
        extra_select = ", " + ", ".join(extra_select)
    else:
        extra_select = ""

    if len(extra_where) > 0:
        extra_where = " AND " + " AND ".join(extra_where)
    else:
        extra_where = ""

    if len(extra_groupby) > 0:
        extra_groupby = ", " + ", ".join(extra_groupby)
    else:
        extra_groupby = ""

    if len(extra_having) > 0:
        extra_having = " HAVING " + " AND ".join(extra_having)
    else:
        extra_having = ""

    return """SELECT
        ps.srvid, qualid, ps.queryid, query, ps.dbid,
        to_json(quals) AS quals,
        sum(execution_count) AS execution_count,
        sum(occurences) AS occurences,
        (sum(nbfiltered) / sum(occurences)) AS avg_filter,
        {filter_ratio} AS filter_ratio
        {extra_select}
        FROM
        {base_query}
        JOIN {{powa}}.powa_statements ps USING(queryid, srvid)
        {extra_join}
        WHERE h.srvid = %(server)s
        {extra_where}
        GROUP BY ps.srvid, qualid, ps.queryid, ps.dbid, ps.query, quals
        {extra_groupby}
        {extra_having}""".format(
        filter_ratio=QUALSTAT_FILTER_RATIO,
        extra_select=extra_select,
        base_query=base_query,
        extra_join=extra_join,
        extra_where=extra_where,
        extra_groupby=extra_groupby,
        extra_having=extra_having,
    )


TEXTUAL_INDEX_QUERY = """
SELECT 'CREATE INDEX idx_' || q.relid || '_' || array_to_string(attnames, '_')
    || ' ON ' || nspname || '.' || q.relid
    || ' USING ' || idxtype || ' (' || array_to_string(attnames, ', ') || ')'
    AS index_ddl
FROM (SELECT t.nspname,
    t.relid,
    t.attnames,
    unnest(t.possible_types) AS idxtype
    FROM (
        SELECT nl.nspname AS nspname,
            qs.relid::regclass AS relid,
            array_agg(DISTINCT attnames.attnames) AS attnames,
            array_agg(DISTINCT pg_am.amname) AS possible_types,
            array_agg(DISTINCT attnum.attnum) AS attnums
        FROM (
            VALUES (:relid, (:attnums)::smallint[], (:indexam))
        ) as qs(relid, attnums, indexam)
        LEFT JOIN (
            pg_class cl
            JOIN pg_namespace nl ON nl.oid = cl.relnamespace
        ) ON cl.oid = qs.relid
        JOIN pg_am  ON pg_am.amname = qs.indexam
            AND pg_am.amname <> 'hash',
        LATERAL (
            SELECT pg_attribute.attname AS attnames
            FROM pg_attribute
            JOIN unnest(qs.attnums) a(a) ON a.a = pg_attribute.attnum
                AND pg_attribute.attrelid = qs.relid
            ORDER BY pg_attribute.attnum
        ) attnames,
        LATERAL unnest(qs.attnums) attnum(attnum)
       WHERE NOT (EXISTS (
           SELECT 1
           FROM pg_index i
           WHERE i.indrelid = qs.relid AND (
             (i.indkey::smallint[])[0:array_length(qs.attnums, 1) - 1]
                 @> qs.attnums
             OR qs.attnums
                 @> (i.indkey::smallint[])[0:array_length(i.indkey, 1) + 1]
             AND i.indisunique))
       )
       GROUP BY nl.nspname, qs.relid
    ) t
    GROUP BY t.nspname, t.relid, t.attnames, t.possible_types
) q
"""


def get_config_changes(restrict_database=False):
    restrict_db = ""
    if restrict_database:
        restrict_db = "AND (d.datname = %(database)s OR h.setdatabase = 0)"

    sql = """SELECT * FROM
(
  WITH src AS (
    select ts, name,
    lag(setting_pretty) OVER (PARTITION BY name ORDER BY ts) AS prev_val,
    setting_pretty AS new_val,
    lag(is_dropped) OVER (PARTITION BY name ORDER BY ts) AS prev_is_dropped,
    is_dropped as is_dropped
    FROM {{pg_track_settings}}.pg_track_settings_history h
    WHERE srvid = %(server)s
    AND ts <= %(to)s
  )
  SELECT extract("epoch" FROM ts) AS ts, 'global' AS kind,
  json_build_object(
    'name', name,
    'prev_val', prev_val,
    'new_val', new_val,
    'prev_is_dropped', coalesce(prev_is_dropped, true),
    'is_dropped', is_dropped
  ) AS data
  FROM src
  WHERE ts >= %(from)s AND ts <= %(to)s
) AS global

UNION ALL

SELECT * FROM
(
  WITH src AS (
    select ts, name,
    lag(setting) OVER (PARTITION BY name, setdatabase, setrole ORDER BY ts) AS prev_val,
    setting AS new_val,
    lag(is_dropped) OVER (PARTITION BY name, setdatabase, setrole ORDER BY ts) AS prev_is_dropped,
    is_dropped as is_dropped,
    d.datname,
    h.setrole
    FROM {{pg_track_settings}}.pg_track_db_role_settings_history h
    LEFT JOIN {{powa}}.powa_databases d
      ON d.srvid = h.srvid
      AND d.oid = h.setdatabase
    WHERE h.srvid = %(server)s
    {restrict_db}
    AND ts <= %(to)s
  )
  SELECT extract("epoch" FROM ts) AS ts, 'rds' AS kind,
  json_build_object(
    'name', name,
    'prev_val', prev_val,
    'new_val', new_val,
    'prev_is_dropped', coalesce(prev_is_dropped, true),
    'is_dropped', is_dropped,
    'datname', datname,
    'setrole', setrole
  ) AS data
  FROM src
  WHERE ts >= %(from)s AND ts <= %(to)s
) AS rds

UNION ALL

SELECT extract("epoch" FROM ts) AS ts, 'reboot' AS kind,
NULL AS data
FROM {{pg_track_settings}}.pg_reboot AS r
WHERE r.srvid = %(server)s
AND r.ts>= %(from)s
AND r.ts <= %(to)s
ORDER BY ts""".format(restrict_db=restrict_db)

    return sql



"""
Functions to generate the queries used in the various graphs components
"""


def wal_to_num(walname):
    """
    Extracts the sequence number from the given WAL file name, similarly to
    pg_split_walfile_name().  It's assuming a 16MB wal_segment_size.
    """
    return """(('x' || substr({walname}, 9, 8))::bit(32)::bigint
 * '4294967296'::bigint / 16777216
 + ('x' || substr({walname}, 17, 8))::bit(32)::bigint)""".format(
        walname=walname
    )


class Biggest(object):
    def __init__(self, base_columns, order_by):
        if type(base_columns) is str:
            base_columns = [base_columns]
        if type(order_by) is str:
            order_by = [order_by]

        self.base_columns = base_columns
        self.order_by = order_by

    def __call__(self, var, minval=0, label=None):
        label = label or var

        sql = (
            "greatest(lead({var})"
            " OVER (PARTITION BY {partitionby} ORDER BY {orderby})"
            " - {var},"
            " {minval})"
            " AS {alias}".format(
                var=var,
                orderby=", ".join(self.order_by),
                partitionby=", ".join(self.base_columns),
                minval=minval,
                alias=label,
            )
        )
        return sql


class Biggestsum(object):
    def __init__(self, base_columns, order_by):
        if type(base_columns) is str:
            base_columns = [base_columns]
        if type(order_by) is str:
            order_by = [order_by]

        self.base_columns = base_columns
        self.order_by = order_by

    def __call__(self, var, minval=0, label=None):
        label = label or var

        sql = (
            "greatest(lead(sum({var}))"
            " OVER (PARTITION BY {partitionby} ORDER BY {orderby})"
            " - sum({var}),"
            " {minval})"
            " AS {alias}".format(
                var=var,
                orderby=", ".join(self.order_by),
                partitionby=", ".join(self.base_columns),
                minval=minval,
                alias=label,
            )
        )
        return sql


BASE_QUERY_SAMPLE_DB = """(
  SELECT d.srvid, d.datname, base.* FROM {powa}.powa_databases d,
  LATERAL (
    SELECT *
    FROM (
      SELECT
      row_number() OVER (
        PARTITION BY dbid ORDER BY statements_history.ts
      ) AS number,
      count(*) OVER (PARTITION BY dbid) AS total,
      *
      FROM (
        SELECT dbid, (unnested.records).*
        FROM (
          SELECT psh.dbid, psh.coalesce_range, unnest(records) AS records
          FROM {powa}.powa_statements_history_db psh
          WHERE coalesce_range && tstzrange(%(from)s, %(to)s,'[]')
          AND psh.dbid = d.oid
          AND psh.srvid = d.srvid
          AND psh.srvid = %(server)s
        ) AS unnested
        WHERE (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        UNION ALL
        SELECT dbid, (record).*
        FROM {powa}.powa_statements_history_current_db
        WHERE  (record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        AND dbid = d.oid
        AND srvid = d.srvid
        AND srvid = %(server)s
      ) AS statements_history
    ) AS sh
    WHERE number %% ( int8larger((total)/(%(samples)s+1),1) ) = 0
  ) AS base
  WHERE srvid = %(server)s
) AS by_db"""


BASE_QUERY_SAMPLE = """(
  SELECT powa_statements.srvid, datname, dbid, queryid, userid, base.*
  FROM {powa}.powa_statements
  JOIN {powa}.powa_databases ON powa_databases.oid = powa_statements.dbid
   AND powa_databases.srvid = powa_statements.srvid,
  LATERAL (
      SELECT *
      FROM (SELECT
          row_number() OVER (
            PARTITION BY queryid ORDER BY statements_history.ts
          ) AS number,
          count(*) OVER (PARTITION BY queryid) AS total,
          *
          FROM (
              SELECT (unnested.records).*
              FROM (
                  SELECT psh.queryid, psh.coalesce_range,
                    unnest(records) AS records
                  FROM {powa}.powa_statements_history psh
                  WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
                  AND psh.queryid = powa_statements.queryid
                  AND psh.userid = powa_statements.userid
                  AND psh.srvid = %(server)s
              ) AS unnested
              WHERE  (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')
              UNION ALL
              SELECT (record).*
              FROM {powa}.powa_statements_history_current phc
              WHERE (record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
              AND phc.queryid = powa_statements.queryid
              AND phc.userid = powa_statements.userid
              AND phc.srvid = %(server)s
          ) AS statements_history
      ) AS sh
      WHERE number %% ( int8larger((total)/(%(samples)s+1),1) ) = 0
  ) AS base
  WHERE powa_statements.srvid = %(server)s
) AS by_query

"""


def powa_getstatdata_sample(mode, predicates=[]):
    """
    predicates is an optional array of plain-text predicates.
    """
    if mode == "db":
        base_query = BASE_QUERY_SAMPLE_DB
        base_columns = ["srvid", "dbid"]

    elif mode == "query":
        base_query = BASE_QUERY_SAMPLE
        base_columns = ["srvid", "dbid", "queryid", "userid"]

    biggest = Biggest(base_columns, "ts")
    biggestsum = Biggestsum(base_columns, "ts")

    where = " AND ".join(predicates)
    if where != "":
        where = " AND " + where

    cols = base_columns + [
        "ts",
        biggest("ts", "'0 s'", "mesure_interval"),
        biggestsum("calls"),
        biggestsum("total_plan_time", label="plantime"),
        biggestsum("total_exec_time", label="runtime"),
        biggestsum("rows"),
        biggestsum("shared_blks_read"),
        biggestsum("shared_blks_hit"),
        biggestsum("shared_blks_dirtied"),
        biggestsum("shared_blks_written"),
        biggestsum("local_blks_read"),
        biggestsum("local_blks_hit"),
        biggestsum("local_blks_dirtied"),
        biggestsum("local_blks_written"),
        biggestsum("temp_blks_read"),
        biggestsum("temp_blks_written"),
        biggestsum("shared_blk_read_time"),
        biggestsum("shared_blk_write_time"),
        biggestsum("local_blk_read_time"),
        biggestsum("local_blk_write_time"),
        biggestsum("temp_blk_read_time"),
        biggestsum("temp_blk_write_time"),
        biggestsum("wal_records"),
        biggestsum("wal_fpi"),
        biggestsum("wal_bytes"),
        biggestsum("jit_functions"),
        biggestsum("jit_generation_time"),
        biggestsum("jit_inlining_count"),
        biggestsum("jit_inlining_time"),
        biggestsum("jit_optimization_count"),
        biggestsum("jit_optimization_time"),
        biggestsum("jit_emission_count"),
        biggestsum("jit_emission_time"),
        biggestsum("jit_deform_count"),
        biggestsum("jit_deform_time"),
    ]

    return """SELECT {cols}
    FROM {base_query}
    WHERE srvid = %(server)s
    {where}
    GROUP BY {base_columns}, ts""".format(
        cols=", ".join(cols),
        base_query=base_query,
        where=where,
        base_columns=", ".join(base_columns),
    )


BASE_QUERY_KCACHE_SAMPLE_DB = """
        {powa}.powa_databases d,
        LATERAL (
            SELECT *
            FROM (
                SELECT row_number() OVER (ORDER BY kmbq.ts) AS number,
                    count(*) OVER () as total,
                        *
                FROM (
                    SELECT km.ts,
                    sum(km.plan_reads + km.exec_reads) AS reads,
                    sum(km.plan_writes + km.exec_writes) AS writes,
                    sum(km.plan_user_time + km.exec_user_time) AS user_time,
                    sum(km.plan_system_time + km.exec_system_time) AS system_time,
                    sum(km.plan_minflts + km.exec_minflts) AS minflts,
                    sum(km.plan_majflts + km.exec_majflts) AS majflts,
                    -- not maintained on GNU/Linux, and not available on Windows
                    -- sum(km.plan_nswaps + km.exec_nswaps) AS nswaps,
                    -- sum(km.plan_msgsnds + km.exec_msgsnds) AS msgsnds,
                    -- sum(km.plan_msgrcvs + km.exec_msgrcvs) AS msgrcvs,
                    -- sum(km.plan_nsignals + km.exec_nsignals) AS nsignals,
                    sum(km.plan_nvcsws + km.exec_nvcsws) AS nvcsws,
                    sum(km.plan_nivcsws + km.exec_nivcsws) AS nivcsws
                    FROM (
                        SELECT * FROM (
                            SELECT (unnest(metrics)).*
                            FROM {powa}.powa_kcache_metrics_db kmd
                            WHERE kmd.srvid = d.srvid
                            AND kmd.dbid = d.oid
                            AND kmd.coalesce_range &&
                                tstzrange(%(from)s, %(to)s, '[]')
                        ) his
                        WHERE his.ts <@ tstzrange(%(from)s, %(to)s, '[]')
                        UNION ALL
                        SELECT (metrics).*
                        FROM {powa}.powa_kcache_metrics_current_db kmcd
                        WHERE kmcd.srvid = d.srvid
                        AND kmcd.dbid = d.oid
                        AND (metrics).ts <@ tstzrange(%(from)s, %(to)s, '[]')
                    ) km
                    GROUP BY km.ts
                ) kmbq
            ) kmn
        WHERE kmn.number %% (int8larger(total/(%(samples)s+1),1) ) = 0
        ) kcache
"""


BASE_QUERY_KCACHE_SAMPLE = """
        {powa}.powa_statements s JOIN {powa}.powa_databases d
            ON d.oid = s.dbid AND d.srvid = s.srvid
            AND s.srvid = %(server)s,
        LATERAL (
            SELECT *
            FROM (
                SELECT row_number() OVER (ORDER BY kmbq.ts) AS number,
                    count(*) OVER () as total,
                        *
                FROM (
                    SELECT km.ts,
                    sum(km.plan_reads + km.exec_reads) AS reads,
                    sum(km.plan_writes + km.exec_writes) AS writes,
                    sum(km.plan_user_time + km.exec_user_time) AS user_time,
                    sum(km.plan_system_time + km.exec_system_time) AS system_time,
                    sum(km.plan_minflts + km.exec_minflts) AS minflts,
                    sum(km.plan_majflts + km.exec_majflts) AS majflts,
                    -- not maintained on GNU/Linux, and not available on Windows
                    -- sum(km.plan_nswaps + km.exec_nswaps) AS nswaps,
                    -- sum(km.plan_msgsnds + km.exec_msgsnds) AS msgsnds,
                    -- sum(km.plan_msgrcvs + km.exec_msgrcvs) AS msgrcvs,
                    -- sum(km.plan_nsignals + km.exec_nsignals) AS nsignals,
                    sum(km.plan_nvcsws + km.exec_nvcsws) AS nvcsws,
                    sum(km.plan_nivcsws + km.exec_nivcsws) AS nivcsws
                    FROM (
                        SELECT * FROM (
                            SELECT (unnest(metrics)).*
                            FROM {powa}.powa_kcache_metrics km
                            WHERE km.srvid = s.srvid
                            AND km.queryid = s.queryid
                            AND km.dbid = s.dbid
                            AND km.coalesce_range &&
                                tstzrange(%(from)s, %(to)s, '[]')
                        ) his
                        WHERE his.ts <@ tstzrange(%(from)s, %(to)s, '[]')
                        UNION ALL
                        SELECT (metrics).*
                        FROM {powa}.powa_kcache_metrics_current kmc
                        WHERE kmc.srvid = s.srvid
                        AND kmc.queryid = s.queryid
                        AND kmc.dbid = s.dbid
                        AND (metrics).ts <@ tstzrange(%(from)s, %(to)s, '[]')
                    ) km
                    GROUP BY km.ts
                ) kmbq
            ) kmn
        WHERE kmn.number %% (int8larger(total/(%(samples)s+1),1) ) = 0
        ) kcache
"""


def kcache_getstatdata_sample(mode, predicates=[]):
    """
    predicates is an optional array of plain-text predicates.
    """
    if mode == "db":
        base_query = BASE_QUERY_KCACHE_SAMPLE_DB
        base_columns = ["d.oid AS dbid", "srvid, datname"]
        groupby_columns = "d.oid, srvid, datname"
    elif mode == "query":
        base_query = BASE_QUERY_KCACHE_SAMPLE
        base_columns = [
            "d.oid AS dbid",
            "d.srvid",
            "datname",
            "queryid",
            "userid",
        ]
        groupby_columns = "d.oid, d.srvid, datname, queryid, userid"

    biggestsum = Biggestsum(groupby_columns, "ts")

    where = " AND ".join(predicates)
    if where != "":
        where = " AND " + where

    base_columns.extend(
        [
            "ts",
            biggestsum("reads"),
            biggestsum("writes"),
            biggestsum("user_time"),
            biggestsum("system_time"),
            biggestsum("minflts"),
            biggestsum("majflts"),
            # not maintained on GNU/Linux, and not available on Windows
            # biggestsum("nswaps"),
            # biggestsum("msgsnds"),
            # biggestsum("msgrcvs"),
            # biggestsum("nsignals"),
            biggestsum("nvcsws"),
            biggestsum("nivcsws"),
        ]
    )

    return """SELECT {base_columns}
    FROM {base_query}
    WHERE d.srvid = %(server)s
    {where}
    GROUP BY {groupby_columns}, ts""".format(
        base_columns=", ".join(base_columns),
        groupby_columns=groupby_columns,
        where=where,
        base_query=base_query,
    )


BASE_QUERY_WAIT_SAMPLE_DB = """(
  SELECT d.oid AS dbid, datname, base.*
  FROM {powa}.powa_databases d,
  LATERAL (
    SELECT *
    FROM (SELECT
      row_number() OVER (
        PARTITION BY dbid ORDER BY waits_history.ts
      ) AS number,
      count(*) OVER (PARTITION BY dbid) AS total,
      srvid,
      ts,
      -- pg 96 columns (bufferpin and lock are included in pg 10+)
      sum(count) FILTER
        (WHERE event_type = 'LWLockNamed') as count_lwlocknamed,
      sum(count) FILTER
        (WHERE event_type = 'LWLockTranche') as count_lwlocktranche,
      -- pg 10+ columns
      sum(count) FILTER (WHERE event_type = 'LWLock') as count_lwlock,
      sum(count) FILTER (WHERE event_type = 'Lock') as count_lock,
      sum(count) FILTER (WHERE event_type = 'BufferPin') as count_bufferpin,
      sum(count) FILTER (WHERE event_type = 'Activity') as count_activity,
      sum(count) FILTER (WHERE event_type = 'Client') as count_client,
      sum(count) FILTER (WHERE event_type = 'Extension') as count_extension,
      sum(count) FILTER (WHERE event_type = 'IPC') as count_ipc,
      sum(count) FILTER (WHERE event_type = 'Timeout') as count_timeout,
      sum(count) FILTER (WHERE event_type = 'IO') as count_io
      FROM (
        SELECT srvid, dbid, event_type, (unnested.records).ts,
          sum((unnested.records).count) AS count
        FROM (
          SELECT wsh.srvid, wsh.dbid, wsh.coalesce_range, event_type,
              unnest(records) AS records
          FROM {powa}.powa_wait_sampling_history_db wsh
          WHERE coalesce_range && tstzrange(%(from)s, %(to)s,'[]')
          AND wsh.dbid = d.oid
          AND wsh.srvid = %(server)s
        ) AS unnested
        WHERE (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        GROUP BY unnested.srvid, unnested.dbid, unnested.event_type,
          (unnested.records).ts
        UNION ALL
        SELECT wshc.srvid, wshc.dbid, event_type, (wshc.record).ts,
            sum((wshc.record).count) AS count
        FROM {powa}.powa_wait_sampling_history_current_db wshc
        WHERE (wshc.record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        AND wshc.dbid = d.oid
        AND wshc.srvid = %(server)s
        GROUP BY wshc.srvid, wshc.dbid, wshc.event_type, (wshc.record).ts
      ) AS waits_history
      GROUP BY ts, srvid, dbid
    ) AS wh
    WHERE number %% ( int8larger((total)/(%(samples)s+1),1) ) = 0
    AND wh.srvid = d.srvid
  ) AS base
  WHERE d.srvid = %(server)s
) AS by_db
"""


BASE_QUERY_WAIT_SAMPLE = """(
  SELECT d.srvid, datname, dbid, queryid, base.*
  FROM {powa}.powa_statements s
  JOIN {powa}.powa_databases d ON d.oid = s.dbid
      AND d.srvid = s.srvid,
  LATERAL (
    SELECT *
    FROM (SELECT
      row_number() OVER (
        PARTITION BY queryid ORDER BY waits_history.ts
      ) AS number,
      count(*) OVER (PARTITION BY queryid) AS total,
      ts,
      -- pg 96 columns (bufferpin and lock are included in pg 10+)
      sum(count) FILTER
        (WHERE event_type = 'LWLockNamed') AS count_lwlocknamed,
      sum(count) FILTER
        (WHERE event_type = 'LWLockTranche') AS count_lwlocktranche,
      -- pg 10+ columns
      sum(count) FILTER (WHERE event_type = 'LWLock') AS count_lwlock,
      sum(count) FILTER (WHERE event_type = 'Lock') AS count_lock,
      sum(count) FILTER (WHERE event_type = 'BufferPin') AS count_bufferpin,
      sum(count) FILTER (WHERE event_type = 'Activity') AS count_activity,
      sum(count) FILTER (WHERE event_type = 'Client') AS count_client,
      sum(count) FILTER (WHERE event_type = 'Extension') AS count_extension,
      sum(count) FILTER (WHERE event_type = 'IPC') AS count_ipc,
      sum(count) FILTER (WHERE event_type = 'Timeout') AS count_timeout,
      sum(count) FILTER (WHERE event_type = 'IO') AS count_io
      FROM (
        SELECT unnested.event_type, (unnested.records).ts,
          sum((unnested.records).count) AS count
        FROM (
          SELECT coalesce_range, event_type,
              unnest(records) AS records
          FROM {powa}.powa_wait_sampling_history wsh
          WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
          AND wsh.queryid = s.queryid
          AND wsh.srvid = %(server)s
        ) AS unnested
        WHERE (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        GROUP BY unnested.event_type, (unnested.records).ts
        UNION ALL
        SELECT wshc.event_type, (wshc.record).ts,
          sum((wshc.record).count) AS count
        FROM {powa}.powa_wait_sampling_history_current wshc
        WHERE (wshc.record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        AND wshc.queryid = s.queryid
        AND wshc.srvid = %(server)s
        GROUP BY wshc.srvid, wshc.event_type, (wshc.record).ts
      ) AS waits_history
      GROUP BY waits_history.ts
    ) AS sh
    WHERE number %% ( int8larger((total)/(%(samples)s+1),1) ) = 0
  ) AS base
  WHERE d.srvid = %(server)s
) AS by_query
"""


# Note about xid calculations: we store 32b xids and those wraparounds, so we
# have to account for that.  At the same time, cur_txid is retrieved before
# executing the query fetching pg_stat_activity information.  On some busy
# system, this could take quite some time and nothing prevents xid from being
# assigned during that time, so to avoid returning negative number we
# arbitrarily choose 100k transactions as a cutoff point to distinguish
# wraparound vs really negative numbers.
# For any number less than -100k, we assume this is because cur_txid
# wraparound and the xid didn't, in which case we compute the actual value,
# knowing that the first 3 xid are reserved and can never be assigned, and that
# the highest transaction id is 2^32.
def BASE_QUERY_PGSA_SAMPLE(per_db=False):
    if per_db:
        extra = """JOIN {powa}.powa_databases d
            ON d.oid = pgsa_history.datid and d.srvid = pgsa_history.srvid
        WHERE d.datname = %(database)s"""
    else:
        extra = ""

    # We use dense_rank() as we need ALL the records for a specific ts
    return """
    (
      SELECT *, max(number) OVER () AS total
      FROM (
        SELECT pgsa_history.srvid,
          dense_rank() OVER (ORDER BY pgsa_history.ts) AS number,
          ts,
          datid,
          cur_txid,
          backend_xid,
          backend_xmin,
          backend_start,
          xact_start,
          query_start,
          state,
          leader_pid
        FROM (
          SELECT *
          FROM (
            SELECT srvid, (unnest(records)).*
            FROM {{powa}}.powa_stat_activity_history pgsah
            WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
            AND pgsah.srvid = %(server)s
          ) AS unnested
          WHERE ts <@ tstzrange(%(from)s, %(to)s, '[]')
          AND backend_type <> 'autovacuum worker'
          UNION ALL
          SELECT srvid, (record).*
          FROM {{powa}}.powa_stat_activity_history_current pgsac
          WHERE (pgsac.record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
          AND pgsac.srvid = %(server)s
          AND (record).backend_type <> 'autovacuum worker'
          ) AS pgsa_history
          {extra}
      ) AS ranking
    ) AS pgsa
    WHERE number %% ( int8larger((total)/(%(samples)s+1),1) ) = 0
""".format(extra=extra)


BASE_QUERY_ARCHIVER_SAMPLE = """
    (SELECT srvid,
      row_number() OVER (ORDER BY arc_history.ts) AS number,
      count(*) OVER () AS total,
      ts,
      current_wal,
      archived_count,
      last_archived_wal,
      last_archived_time,
      failed_count,
      last_failed_wal,
      last_failed_time
      FROM (
        SELECT *
        FROM (
          SELECT srvid, (unnest(records)).*
          FROM {powa}.powa_stat_archiver_history arch
          WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
          AND arch.srvid = %(server)s
        ) AS unnested
        WHERE ts <@ tstzrange(%(from)s, %(to)s, '[]')
        UNION ALL
        SELECT srvid, (record).*
        FROM {powa}.powa_stat_archiver_history_current arcc
        WHERE (arcc.record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        AND arcc.srvid = %(server)s
      ) AS arc_history
    ) AS arc
    WHERE number %% ( int8larger((total)/(%(samples)s+1),1) ) = 0
"""


BASE_QUERY_BGWRITER_SAMPLE = """
    (SELECT srvid,
      row_number() OVER (ORDER BY bgw_history.ts) AS number,
      count(*) OVER () AS total,
      ts,
      buffers_clean,
      maxwritten_clean,
      buffers_backend,
      buffers_backend_fsync,
      buffers_alloc
      FROM (
        SELECT *
        FROM (
          SELECT srvid, (unnest(records)).*
          FROM {powa}.powa_stat_bgwriter_history bgwh
          WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
          AND bgwh.srvid = %(server)s
        ) AS unnested
        WHERE ts <@ tstzrange(%(from)s, %(to)s, '[]')
        UNION ALL
        SELECT srvid, (record).*
        FROM {powa}.powa_stat_bgwriter_history_current bgwc
        WHERE (bgwc.record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        AND bgwc.srvid = %(server)s
      ) AS bgw_history
    ) AS bgw
    WHERE number %% ( int8larger((total)/(%(samples)s+1),1) ) = 0
"""


BASE_QUERY_CHECKPOINTER_SAMPLE = """
    (SELECT srvid,
      row_number() OVER (ORDER BY cpt_history.ts) AS number,
      count(*) OVER () AS total,
      ts,
      num_timed,
      num_requested,
      write_time,
      sync_time,
      buffers_written
      FROM (
        SELECT *
        FROM (
          SELECT srvid, (unnest(records)).*
          FROM {powa}.powa_stat_checkpointer_history cpth
          WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
          AND cpth.srvid = %(server)s
        ) AS unnested
        WHERE ts <@ tstzrange(%(from)s, %(to)s, '[]')
        UNION ALL
        SELECT srvid, (record).*
        FROM {powa}.powa_stat_checkpointer_history_current cptc
        WHERE (cptc.record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        AND cptc.srvid = %(server)s
      ) AS cpt_history
    ) AS cpt
    WHERE number %% ( int8larger((total)/(%(samples)s+1),1) ) = 0
"""


def BASE_QUERY_DATABASE_SAMPLE(per_db=False):
    if per_db:
        extra = """JOIN {powa}.powa_databases d
            ON d.oid = psd_history.datid and d.srvid = psd_history.srvid
        WHERE d.datname = %(database)s"""
    else:
        extra = ""

    return """
    (SELECT psd_history.srvid,
      row_number() OVER (ORDER BY psd_history.ts) AS number,
      count(*) OVER () AS total,
      ts,
      sum(numbackends) AS numbackends,
      sum(xact_commit) AS xact_commit,
      sum(xact_rollback) AS xact_rollback,
      sum(blks_read) AS blks_read,
      sum(blks_hit) AS blks_hit,
      sum(tup_returned) AS tup_returned,
      sum(tup_fetched) AS tup_fetched,
      sum(tup_inserted) AS tup_inserted,
      sum(tup_updated) AS tup_updated,
      sum(tup_deleted) AS tup_deleted,
      sum(conflicts) AS conflicts,
      sum(temp_files) AS temp_files,
      sum(temp_bytes) AS temp_bytes,
      sum(deadlocks) AS deadlocks,
      sum(checksum_failures) AS checksum_failures,
      max(checksum_last_failure) AS checksum_last_failure,
      sum(blk_read_time) AS blk_read_time,
      sum(blk_write_time) AS blk_write_time,
      sum(session_time) AS session_time,
      sum(active_time) AS active_time,
      sum(idle_in_transaction_time) AS idle_in_transaction_time,
      sum(sessions) AS sessions,
      sum(sessions_abandoned) AS sessions_abandoned,
      sum(sessions_fatal) AS sessions_fatal,
      sum(sessions_killed) AS sessions_killed,
      max(stats_reset) AS stats_reset
      FROM (
        SELECT *
        FROM (
          SELECT srvid, datid, (unnest(records)).*
          FROM {{powa}}.powa_stat_database_history psdh
          WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
          AND psdh.srvid = %(server)s
        ) AS unnested
        WHERE ts <@ tstzrange(%(from)s, %(to)s, '[]')
        UNION ALL
        SELECT srvid, datid, (record).*
        FROM {{powa}}.powa_stat_database_history_current psdc
        WHERE (psdc.record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        AND psdc.srvid = %(server)s
      ) AS psd_history
      {extra}
      GROUP BY psd_history.srvid, psd_history.ts
    ) AS psd
    WHERE number %% ( int8larger((total)/(%(samples)s+1),1) ) = 0
""".format(extra=extra)


def BASE_QUERY_DATABASE_CONFLICTS_SAMPLE(per_db=False):
    if per_db:
        extra = """JOIN {powa}.powa_databases d
            ON d.oid = psd_history.datid and d.srvid = psd_history.srvid
        WHERE d.datname = %(database)s"""
    else:
        extra = ""

    return """
    (SELECT psdc_history.srvid,
      row_number() OVER (ORDER BY psdc_history.ts) AS number,
      count(*) OVER () AS total,
      ts,
      sum(confl_tablespace) AS confl_tablespace,
      sum(confl_lock) AS confl_lock,
      sum(confl_snapshot) AS confl_snapshot,
      sum(confl_bufferpin) AS confl_bufferpin,
      sum(confl_deadlock) AS confl_deadlock,
      sum(confl_active_logicalslot) AS confl_active_logicalslot
      FROM (
        SELECT *
        FROM (
          SELECT srvid, (unnest(records)).*
          FROM {{powa}}.powa_stat_database_conflicts_history psdch
          WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
          AND psdch.srvid = %(server)s
        ) AS unnested
        WHERE ts <@ tstzrange(%(from)s, %(to)s, '[]')
        UNION ALL
        SELECT srvid, (record).*
        FROM {{powa}}.powa_stat_database_conflicts_history_current psdcc
        WHERE (psdcc.record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        AND psdcc.srvid = %(server)s
      ) AS psdc_history
      {extra}
      GROUP BY psdc_history.srvid, psdc_history.ts
    ) AS psdc
    WHERE number %% ( int8larger((total)/(%(samples)s+1),1) ) = 0
""".format(extra=extra)


# We use dense_rank() as we need ALL the records for a specific ts.  Caller
# will group the data per backend_type, object or other as needed
BASE_QUERY_IO_SAMPLE = """
    (SELECT srvid,
      --dense_rank() OVER (ORDER BY ts, backend_type, object, context) AS number,
      dense_rank() OVER (ORDER BY ts) AS number,
      count(*) OVER (PARTITION BY ts) AS num_per_window,
      count(*) OVER () AS total,
      ts,
      backend_type, object, context,
      reads * op_bytes AS reads, read_time,
      writes * op_bytes AS writes, write_time,
      writebacks * op_bytes AS writebacks, writeback_time,
      extends * op_bytes AS extends, extend_time,
      hits * op_bytes AS hits,
      evictions * op_bytes AS evictions,
      reuses * op_bytes AS reuses,
      fsyncs * op_bytes AS fsyncs, fsync_time AS fsync_time
      FROM (
        SELECT *
        FROM (
          SELECT srvid, backend_type, object, context, (unnest(records)).*
          FROM {powa}.powa_stat_io_history ioh
          WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
          AND ioh.srvid = %(server)s
        ) AS unnested
        WHERE ts <@ tstzrange(%(from)s, %(to)s, '[]')
        UNION ALL
        SELECT srvid, backend_type, object, context, (record).*
        FROM {powa}.powa_stat_io_history_current ioc
        WHERE (ioc.record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        AND ioc.srvid = %(server)s
      ) AS io_history
    ) AS io
    WHERE number %% ( int8larger((total / num_per_window)/(%(samples)s+1),1) ) = 0
"""


BASE_QUERY_REPLICATION_SAMPLE = """
    (SELECT srvid,
      row_number() OVER (ORDER BY psr_history.ts) AS number,
      count(*) OVER () AS total,
      ts,
      rep_current_lsn,
      count(DISTINCT pid) AS nb_repl,
      min(sent_lsn) AS sent_lsn,
      min(write_lsn) AS write_lsn,
      min(flush_lsn) AS flush_lsn,
      min(replay_lsn) AS replay_lsn,
      max(write_lag) AS write_lag,
      max(flush_lag) AS flush_lag,
      max(replay_lag) AS replay_lag,
      count(DISTINCT pid) FILTER (WHERE sync_state = 'async') AS nb_async,
      count(*) FILTER (WHERE slot_type = 'physical' AND active) AS nb_physical_act,
      count(*) FILTER (WHERE slot_type = 'physical' AND NOT active) AS nb_physical_not_act,
      count(*) FILTER (WHERE slot_type = 'logical' AND active) AS nb_logical_act,
      count(*) FILTER (WHERE slot_type = 'logical' AND NOT active) AS nb_logical_not_act
      FROM (
        SELECT statrep.srvid,
          statrep.ts,
          statrep.current_lsn AS rep_current_lsn, statrep.pid, statrep.usename,
          statrep.application_name, statrep.client_addr, statrep.backend_start,
          statrep.backend_xmin, statrep.state, statrep.sent_lsn,
          statrep.write_lsn, statrep.flush_lsn, statrep.replay_lsn,
          statrep.write_lag, statrep.flush_lag, statrep.replay_lag,
          statrep.sync_priority, statrep.sync_state, statrep.reply_time,

          replslot.cur_txid, replslot.current_lsn AS slot_current_lsn,
          replslot.slot_name, replslot.plugin, replslot.slot_type,
          replslot.datoid, replslot.temporary, replslot.active,
          replslot.active_pid, replslot.slot_xmin, replslot.catalog_xmin,
          replslot.restart_lsn, replslot.confirmed_flush_lsn,
          replslot.wal_status, replslot.safe_wal_size, replslot.two_phase,
          replslot.conflicting
        FROM (
          SELECT srvid, (unnest(records)).*
          FROM {powa}.powa_stat_replication_history psrh
          WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
          AND psrh.srvid = %(server)s
        ) AS statrep
        FULL OUTER JOIN (
          SELECT srvid, slot_name, plugin, slot_type, datoid, temporary,
            (unnest(records)).*
          FROM {powa}.powa_replication_slots_history prsh
          WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
          AND prsh.srvid = %(server)s
        ) AS replslot
          ON replslot.srvid = statrep.srvid
          AND replslot.ts = statrep.ts
          AND replslot.active_pid = statrep.pid
        WHERE statrep.ts <@ tstzrange(%(from)s, %(to)s, '[]')
        UNION ALL
        SELECT %(server)s AS srvid,
          (psrc.record).ts,
          (psrc.record).current_lsn AS rep_current_lsn, (psrc.record).pid,
          (psrc.record).usename, (psrc.record).application_name,
          (psrc.record).client_addr, (psrc.record).backend_start,
          (psrc.record).backend_xmin, (psrc.record).state,
          (psrc.record).sent_lsn, (psrc.record).write_lsn,
          (psrc.record).flush_lsn, (psrc.record).replay_lsn,
          (psrc.record).write_lag, (psrc.record).flush_lag,
          (psrc.record).replay_lag, (psrc.record).sync_priority,
          (psrc.record).sync_state, (psrc.record).reply_time,

          (prsc.record).cur_txid,
          (prsc.record).current_lsn AS slot_current_lsn,
          prsc.slot_name, prsc.plugin,
          prsc.slot_type, prsc.datoid,
          prsc.temporary, (prsc.record).active,
          (prsc.record).active_pid, (prsc.record).slot_xmin,
          (prsc.record).catalog_xmin, (prsc.record).restart_lsn,
          (prsc.record).confirmed_flush_lsn, (prsc.record).wal_status,
          (prsc.record).safe_wal_size, (prsc.record).two_phase,
          (prsc.record).conflicting
        FROM {powa}.powa_stat_replication_history_current psrc
        FULL OUTER JOIN {powa}.powa_replication_slots_history_current prsc
          ON psrc.srvid = prsc.srvid AND (psrc.record).ts = (prsc.record).ts
            AND  (psrc.record).pid = (prsc.record).active_pid
        WHERE (psrc.record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        AND psrc.srvid = %(server)s
      ) AS psr_history
      GROUP BY psr_history.srvid, psr_history.ts, psr_history.rep_current_lsn
    ) AS psr
    WHERE number %% ( int8larger((total)/(%(samples)s+1),1) ) = 0
"""


# We use dense_rank() as we need ALL the records for a specific ts.  Caller
# will group the data as needed
BASE_QUERY_SLRU_SAMPLE = """
    (SELECT srvid,
      dense_rank() OVER (ORDER BY slru_history.ts) AS number,
      count(*) OVER (PARTITION BY ts) AS num_per_window,
      count(*) OVER () AS total,
      ts,
      name,
      blks_zeroed, blks_hit, blks_read, blks_written,
      blks_exists, flushes, truncates
      FROM (
        SELECT *
        FROM (
          SELECT srvid, name, (unnest(records)).*
          FROM {powa}.powa_stat_slru_history slruh
          WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
          AND slruh.srvid = %(server)s
        ) AS unnested
        WHERE ts <@ tstzrange(%(from)s, %(to)s, '[]')
        UNION ALL
        SELECT srvid, name, (record).*
        FROM {powa}.powa_stat_slru_history_current slruc
        WHERE (slruc.record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        AND slruc.srvid = %(server)s
      ) AS slru_history
    ) AS slru
    WHERE number %% ( int8larger((total / num_per_window)/(%(samples)s+1),1) ) = 0
"""


def BASE_QUERY_SUBSCRIPTION_SAMPLE(subname=None):
    if subname is not None:
        extra = "AND subname = %(subscription)s"
    else:
        extra = ""

    # We use dense_rank() as we need ALL the records for a specific ts.  Caller
    # will group the data as needed.
    return """
    (SELECT srvid,
      dense_rank() OVER (ORDER BY ts) AS number,
      count(*) OVER (PARTITION BY ts) AS num_per_window,
      count(*) OVER () AS total,
      ts,
      subname,
      extract(epoch FROM ((last_msg_receipt_time - last_msg_send_time) * 1000)) AS last_msg_lag,
      latest_end_lsn,
      extract(epoch FROM ((ts - latest_end_time) * 1000)) AS report_lag,
      apply_error_count,
      sync_error_count
      FROM (
        SELECT *
        FROM (
          SELECT *
          FROM (
            SELECT srvid, sh.subid, sh.subname, (unnest(records)).*
            FROM {{powa}}.powa_stat_subscription_history sh
            WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
            AND sh.srvid = %(server)s
            {extra}
            ) AS sub1
          LEFT JOIN (
            SELECT srvid, ssh.subid, (unnest(records)).*
            FROM {{powa}}.powa_stat_subscription_stats_history ssh
            WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
            AND ssh.srvid = %(server)s
            {extra}
          ) AS sub2 USING (srvid, subid, ts)
        ) AS unnested
        WHERE ts <@ tstzrange(%(from)s, %(to)s, '[]')
        UNION ALL
        SELECT *
        FROM (
            SELECT srvid, shc.subid, shc.subname, (record).*
            FROM {{powa}}.powa_stat_subscription_history_current shc
            WHERE (shc.record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
            AND shc.srvid = %(server)s
            {extra}
        ) AS sub1
        LEFT JOIN (
            SELECT srvid, sshc.subid, (record).*
            FROM {{powa}}.powa_stat_subscription_stats_history_current sshc
            WHERE (sshc.record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
            AND sshc.srvid = %(server)s
            {extra}
        ) AS sub2 USING (srvid, subid, ts)
      ) AS subscription_history
    ) AS io
    WHERE number %% ( int8larger((total / num_per_window)/(%(samples)s+1),1) ) = 0
""".format(extra=extra)


BASE_QUERY_WAL_SAMPLE = """
    (SELECT srvid,
      row_number() OVER (ORDER BY wal_history.ts) AS number,
      count(*) OVER () AS total,
      ts,
      wal_records,
      wal_fpi,
      wal_bytes,
      wal_buffers_full,
      wal_write,
      wal_sync,
      wal_write_time,
      wal_sync_time
      FROM (
        SELECT *
        FROM (
          SELECT srvid, (unnest(records)).*
          FROM {powa}.powa_stat_wal_history walh
          WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
          AND walh.srvid = %(server)s
        ) AS unnested
        WHERE ts <@ tstzrange(%(from)s, %(to)s, '[]')
        UNION ALL
        SELECT srvid, (record).*
        FROM {powa}.powa_stat_wal_history_current walc
        WHERE (walc.record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        AND walc.srvid = %(server)s
      ) AS wal_history
    ) AS wal
    WHERE number %% ( int8larger((total)/(%(samples)s+1),1) ) = 0
"""


BASE_QUERY_WAL_RECEIVER_SAMPLE = """
    (SELECT srvid,
      row_number() OVER (ORDER BY wr_history.ts) AS number,
      count(*) OVER () AS total,
      ts,
      slot_name,
      sender_host, sender_port,
      pid, status,
      receive_start_lsn, receive_start_tli,
      last_received_lsn,
      written_lsn, flushed_lsn,
      received_tli,
      last_msg_send_time,
      last_msg_receipt_time,
      latest_end_lsn,
      latest_end_time,
      conninfo
      FROM (
        SELECT *
        FROM (
          SELECT srvid, slot_name, sender_host, sender_port, (unnest(records)).*
          FROM {powa}.powa_stat_wal_receiver_history walh
          WHERE coalesce_range && tstzrange(%(from)s, %(to)s, '[]')
          AND walh.srvid = %(server)s
        ) AS unnested
        WHERE ts <@ tstzrange(%(from)s, %(to)s, '[]')
        UNION ALL
        SELECT srvid, slot_name, sender_host, sender_port, (record).*
        FROM {powa}.powa_stat_wal_receiver_history_current walc
        WHERE (walc.record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        AND walc.srvid = %(server)s
      ) AS wr_history
    ) AS wr
    WHERE number %% ( int8larger((total)/(%(samples)s+1),1) ) = 0
"""


BASE_QUERY_ALL_IDXS_SAMPLE_DB = """(
  SELECT d.srvid, d.datname, base.*
  FROM {powa}.powa_databases d,
  LATERAL (
    SELECT *
    FROM (
      SELECT
      row_number() OVER (PARTITION BY dbid ORDER BY ts) AS number,
      count(*) OVER (PARTITION BY dbid ) AS total,
      ar_history.*
      FROM (
        SELECT dbid, (unnested.records).*
        FROM (
          SELECT ari.dbid, unnest(records) AS records
          FROM {powa}.powa_all_indexes_history_db ari
          WHERE ari.dbid = d.oid
          AND ari.srvid = d.srvid
          AND ari.srvid = %(server)s
        ) AS unnested
        WHERE (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        UNION ALL
        SELECT dbid, (record).*
        FROM {powa}.powa_all_indexes_history_current_db aric
        WHERE (record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        AND aric.dbid = d.oid
        AND aric.srvid = d.srvid
        AND aric.srvid = %(server)s
      ) AS ar_history
    ) AS ari
    WHERE number %% (int8larger(total/(%(samples)s+1),1) ) = 0
  ) AS base
  WHERE srvid = %(server)s
) AS by_db
"""


BASE_QUERY_ALL_IDXS_SAMPLE = """(
  SELECT d.srvid, d.datname,
  base.*
  FROM
  {powa}.powa_databases d,
  LATERAL (
    SELECT *
    FROM (
      SELECT
      row_number() OVER (PARTITION BY dbid, relid ORDER BY ts) AS number,
      count(*) OVER (PARTITION BY dbid, relid ) AS total,
      ar_history.*
      FROM (
        SELECT dbid, relid, (unnested.records).*
        FROM (
          SELECT ari.dbid, relid, unnest(records) AS records
          FROM {powa}.powa_all_indexes_history_db ari
          WHERE ari.dbid = d.oid
          AND ari.srvid = d.srvid
          AND ari.srvid = %(server)s
        ) AS unnested
        WHERE (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        UNION ALL
        SELECT dbid, relid, (record).*
        FROM {powa}.powa_all_indexes_history_current_db aric
        WHERE (record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        AND aric.dbid = d.oid
        AND aric.srvid = d.srvid
        AND aric.srvid = %(server)s
      ) AS ar_history
    ) AS ari
    WHERE number %% (int8larger(total/(%(samples)s+1),1) ) = 0
  ) AS base
  WHERE srvid = %(server)s
) AS by_db
    """


BASE_QUERY_ALL_TBLS_SAMPLE_DB = """(
  SELECT d.srvid, d.datname, base.*
  FROM {powa}.powa_databases d,
  LATERAL (
    SELECT *
    FROM (
      SELECT
      row_number() OVER (PARTITION BY dbid ORDER BY ts) AS number,
      count(*) OVER (PARTITION BY dbid ) AS total,
      ar_history.*
      FROM (
        SELECT dbid, (unnested.records).*
        FROM (
          SELECT arh.dbid, unnest(records) AS records
          FROM {powa}.powa_all_tables_history_db arh
          WHERE arh.dbid = d.oid
          AND arh.srvid = d.srvid
          AND arh.srvid = %(server)s
        ) AS unnested
        WHERE (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        UNION ALL
        SELECT dbid, (record).*
        FROM {powa}.powa_all_tables_history_current_db arhc
        WHERE (record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        AND arhc.dbid = d.oid
        AND arhc.srvid = d.srvid
        AND arhc.srvid = %(server)s
      ) AS ar_history
    ) AS arh
    WHERE number %% (int8larger(total/(%(samples)s+1),1) ) = 0
  ) AS base
  WHERE srvid = %(server)s
) AS by_db
"""


BASE_QUERY_ALL_TBLS_SAMPLE = """(
  SELECT d.srvid, d.datname,
  base.*
  FROM
  {powa}.powa_databases d,
  LATERAL (
    SELECT *
    FROM (
      SELECT
      row_number() OVER (PARTITION BY dbid, relid ORDER BY ts) AS number,
      count(*) OVER (PARTITION BY dbid, relid ) AS total,
      ar_history.*
      FROM (
        SELECT dbid, relid, (unnested.records).*
        FROM (
          SELECT arh.dbid, relid, unnest(records) AS records
          FROM {powa}.powa_all_tables_history_db arh
          WHERE arh.dbid = d.oid
          AND arh.srvid = d.srvid
          AND arh.srvid = %(server)s
        ) AS unnested
        WHERE (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        UNION ALL
        SELECT dbid, relid, (record).*
        FROM {powa}.powa_all_tables_history_current_db arhc
        WHERE (record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        AND arhc.dbid = d.oid
        AND arhc.srvid = d.srvid
        AND arhc.srvid = %(server)s
      ) AS ar_history
    ) AS arh
    WHERE number %% (int8larger(total/(%(samples)s+1),1) ) = 0
  ) AS base
  WHERE srvid = %(server)s
) AS by_db
    """


BASE_QUERY_USER_FCTS_SAMPLE_DB = """(
  SELECT d.srvid, d.datname, base.*
  FROM {powa}.powa_databases d,
  LATERAL (
    SELECT *
    FROM (
      SELECT
      row_number() OVER (PARTITION BY dbid ORDER BY ts) AS number,
      count(*) OVER (PARTITION BY dbid ) AS total,
      af_history.*
      FROM (
        SELECT dbid, (unnested.records).*
        FROM (
          SELECT afh.dbid, unnest(records) AS records
          FROM {powa}.powa_user_functions_history_db afh
          WHERE afh.dbid = d.oid
          AND afh.srvid = d.srvid
          AND afh.srvid = %(server)s
        ) AS unnested
        WHERE (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        UNION ALL
        SELECT dbid, (record).*
        FROM {powa}.powa_user_functions_history_current_db afhc
        WHERE (record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        AND afhc.dbid = d.oid
        AND afhc.srvid = d.srvid
        AND afhc.srvid = %(server)s
      ) AS af_history
    ) AS afh
    WHERE number %% (int8larger(total/(%(samples)s+1),1) ) = 0
  ) AS base
  WHERE srvid = %(server)s
) AS by_db
"""


BASE_QUERY_USER_FCTS_SAMPLE = """(
  SELECT d.srvid, d.datname,
  base.*
  FROM
  {powa}.powa_databases d,
  LATERAL (
    SELECT *
    FROM (
      SELECT
      row_number() OVER (PARTITION BY dbid, funcid ORDER BY ts) AS number,
      count(*) OVER (PARTITION BY dbid, funcid ) AS total,
      af_history.*
      FROM (
        SELECT dbid, funcid, (unnested.records).*
        FROM (
          SELECT afh.dbid, funcid, unnest(records) AS records
          FROM {powa}.powa_user_functions_history_db afh
          WHERE afh.dbid = d.oid
          AND afh.srvid = d.srvid
          AND afh.srvid = %(server)s
        ) AS unnested
        WHERE (records).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        UNION ALL
        SELECT dbid, funcid, (record).*
        FROM {powa}.powa_user_functions_history_current_db afhc
        WHERE (record).ts <@ tstzrange(%(from)s, %(to)s, '[]')
        AND afhc.dbid = d.oid
        AND afhc.srvid = d.srvid
        AND afhc.srvid = %(server)s
      ) AS af_history
    ) AS afh
    WHERE number %% (int8larger(total/(%(samples)s+1),1) ) = 0
  ) AS base
  WHERE srvid = %(server)s
) AS by_db
    """


def powa_getwaitdata_sample(mode, predicates=[]):
    """
    predicates is an optional array of plain-text predicates.
    """
    if mode == "db":
        base_query = BASE_QUERY_WAIT_SAMPLE_DB
        base_columns = ["srvid", "dbid"]

    elif mode == "query":
        base_query = BASE_QUERY_WAIT_SAMPLE
        base_columns = ["srvid", "dbid", "queryid"]

    biggest = Biggest(base_columns, "ts")
    biggestsum = Biggestsum(base_columns, "ts")

    where = " AND ".join(predicates)
    if where != "":
        where = " AND " + where

    all_cols = base_columns + [
        "ts",
        biggest("ts", "'0 s'", "mesure_interval"),
        # pg 96 only columns
        biggestsum("count_lwlocknamed"),
        biggestsum("count_lwlocktranche"),
        # pg 10+ columns
        biggestsum("count_lwlock"),
        biggestsum("count_lock"),
        biggestsum("count_bufferpin"),
        biggestsum("count_activity"),
        biggestsum("count_client"),
        biggestsum("count_extension"),
        biggestsum("count_ipc"),
        biggestsum("count_timeout"),
        biggestsum("count_io"),
    ]

    return """SELECT {all_cols}
    FROM {base_query}
    WHERE srvid = %(server)s
    {where}
    GROUP BY {base_columns}, ts""".format(
        all_cols=", ".join(all_cols),
        base_columns=", ".join(base_columns),
        base_query=base_query,
        where=where,
    )


def powa_get_pgsa_sample(per_db=False):
    base_query = BASE_QUERY_PGSA_SAMPLE(per_db)
    base_columns = ["srvid"]

    def txid_age(field):
        ref = "cur_txid"
        alias = field + "_age"

        return """CASE
        WHEN {ref}::text::bigint - {field}::text::bigint < -100000
          THEN ({ref}::text::bigint - 3) +
            ((4::bigint * 1024 * 1024 * 1024) - {field}::text::bigint)
        WHEN {ref}::text::bigint - {field}::text::bigint <= 0
          THEN 0
        ELSE
          {ref}::text::bigint - {field}::text::bigint
      END AS {alias}""".format(field=field, ref=ref, alias=alias)

    def ts_get_sec(field):
        alias = field + "_age"
        return """extract(epoch FROM (ts - {f})) * 1000 AS {a}""".format(
            f=field, a=alias
        )

    all_cols = base_columns + [
        "ts",
        "datid",
        txid_age("backend_xid"),  # backend_xid_age
        txid_age("backend_xmin"),  # backend_xmin_age
        ts_get_sec("backend_start"),  # backend_start_age
        ts_get_sec("xact_start"),  # xact_start_age
        ts_get_sec("query_start"),  # query_start_age
        "state",
        "leader_pid",
    ]

    return """SELECT {all_cols}
    FROM {base_query}""".format(
        all_cols=", ".join(all_cols), base_query=base_query
    )


def powa_get_archiver_sample():
    base_query = BASE_QUERY_ARCHIVER_SAMPLE
    base_columns = ["srvid"]

    biggest = Biggest(base_columns, "ts")

    all_cols = base_columns + [
        "ts",
        biggest("ts", "'0 s'", "mesure_interval"),
        "current_wal",
        "last_archived_time",
        biggest(wal_to_num("last_archived_wal"), label="nb_arch"),
        "last_failed_time",
        wal_to_num("current_wal")
        + " - "
        + wal_to_num("last_archived_wal")
        + " - 1 AS nb_to_arch",
    ]

    return """SELECT {all_cols}
    FROM {base_query}""".format(
        all_cols=", ".join(all_cols), base_query=base_query
    )


def powa_get_bgwriter_sample():
    base_query = BASE_QUERY_BGWRITER_SAMPLE
    base_columns = ["srvid"]

    biggest = Biggest(base_columns, "ts")
    biggestsum = Biggestsum(base_columns, "ts")

    all_cols = base_columns + [
        "ts",
        biggest("ts", "'0 s'", "mesure_interval"),
        biggestsum("buffers_clean"),
        biggestsum("maxwritten_clean"),
        biggestsum("buffers_backend"),
        biggestsum("buffers_backend_fsync"),
        biggestsum("buffers_alloc"),
    ]

    return """SELECT {all_cols}
    FROM {base_query}
    GROUP BY {base_columns}, ts""".format(
        all_cols=", ".join(all_cols),
        base_columns=", ".join(base_columns),
        base_query=base_query,
    )


def powa_get_checkpointer_sample():
    base_query = BASE_QUERY_CHECKPOINTER_SAMPLE
    base_columns = ["srvid"]

    biggest = Biggest(base_columns, "ts")
    biggestsum = Biggestsum(base_columns, "ts")

    all_cols = base_columns + [
        "ts",
        biggest("ts", "'0 s'", "mesure_interval"),
        biggestsum("num_timed"),
        biggestsum("num_requested"),
        biggestsum("write_time"),
        biggestsum("sync_time"),
        biggestsum("buffers_written"),
    ]

    return """SELECT {all_cols}
    FROM {base_query}
    GROUP BY {base_columns}, ts""".format(
        all_cols=", ".join(all_cols),
        base_columns=", ".join(base_columns),
        base_query=base_query,
    )


def powa_get_database_sample(per_db=False):
    base_query = BASE_QUERY_DATABASE_SAMPLE(per_db)
    base_columns = ["srvid"]

    biggest = Biggest(base_columns, "ts")

    all_cols = base_columns + [
        "ts",
        biggest("ts", "'0 s'", "mesure_interval"),
        biggest("numbackends"),
        biggest("xact_commit"),
        biggest("xact_rollback"),
        biggest("blks_read"),
        biggest("blks_hit"),
        biggest("tup_returned"),
        biggest("tup_fetched"),
        biggest("tup_inserted"),
        biggest("tup_updated"),
        biggest("tup_deleted"),
        biggest("conflicts"),
        biggest("temp_files"),
        biggest("temp_bytes"),
        biggest("deadlocks"),
        biggest("checksum_failures"),
        "checksum_last_failure",
        biggest("blk_read_time"),
        biggest("blk_write_time"),
        biggest("session_time"),
        biggest("active_time"),
        biggest("idle_in_transaction_time"),
        biggest("sessions"),
        biggest("sessions_abandoned"),
        biggest("sessions_fatal"),
        biggest("sessions_killed"),
        "stats_reset",
    ]

    return """SELECT {all_cols}
    FROM {base_query}""".format(
        all_cols=", ".join(all_cols), base_query=base_query
    )


def powa_get_database_conflicts_sample(per_db=False):
    base_query = BASE_QUERY_DATABASE_CONFLICTS_SAMPLE(per_db)
    base_columns = ["srvid"]

    biggest = Biggest(base_columns, "ts")

    all_cols = base_columns + [
        "ts",
        biggest("ts", "'0 s'", "mesure_interval"),
        biggest("confl_tablespace"),
        biggest("confl_lock"),
        biggest("confl_snapshot"),
        biggest("confl_bufferpin"),
        biggest("confl_deadlock"),
        biggest("confl_active_logicalslot"),
    ]

    return """SELECT {all_cols}
    FROM {base_query}""".format(
        all_cols=", ".join(all_cols), base_query=base_query
    )


def powa_get_replication_sample():
    base_query = BASE_QUERY_REPLICATION_SAMPLE
    base_columns = ["srvid"]

    biggest = Biggest(base_columns, "ts")

    all_cols = base_columns + [
        "ts",
        biggest("ts", "'0 s'", "mesure_interval"),
        "rep_current_lsn",
        "nb_repl",
        "sent_lsn",
        "write_lsn",
        "flush_lsn",
        "replay_lsn",
        "write_lag",
        "flush_lag",
        "replay_lag",
        "nb_async",
        "nb_repl - nb_async AS nb_sync",
        "nb_physical_act",
        "nb_physical_not_act",
        "nb_logical_act",
        "nb_logical_not_act",
    ]

    return """SELECT {all_cols}
    FROM {base_query}""".format(
        all_cols=", ".join(all_cols), base_query=base_query
    )


def powa_get_io_sample(qual=None):
    base_query = BASE_QUERY_IO_SAMPLE
    base_columns = ["srvid", "backend_type", "object", "context"]

    biggest = Biggest(base_columns, "ts")

    if qual is not None:
        qual = " AND %s" % qual
    else:
        qual = ""

    all_cols = base_columns + [
        "ts",
        biggest("ts", "'0 s'", "mesure_interval"),
        biggest("reads"),
        biggest("read_time"),
        biggest("writes"),
        biggest("write_time"),
        biggest("writebacks"),
        biggest("writeback_time"),
        biggest("extends"),
        biggest("extend_time"),
        biggest("hits"),
        biggest("evictions"),
        biggest("reuses"),
        biggest("fsyncs"),
        biggest("fsync_time"),
    ]

    return """SELECT {all_cols}
    FROM {base_query}
    {qual}""".format(
        all_cols=", ".join(all_cols), base_query=base_query, qual=qual
    )


def powa_get_slru_sample(qual=None):
    base_query = BASE_QUERY_SLRU_SAMPLE
    base_columns = ["srvid", "name"]

    biggest = Biggest(base_columns, "ts")

    if qual is not None:
        qual = " AND %s" % qual
    else:
        qual = ""

    all_cols = base_columns + [
        "ts",
        biggest("ts", "'0 s'", "mesure_interval"),
        biggest("blks_zeroed"),
        biggest("blks_hit"),
        biggest("blks_read"),
        biggest("blks_written"),
        biggest("blks_exists"),
        biggest("flushes"),
        biggest("truncates"),
    ]

    return """SELECT {all_cols}
    FROM {base_query}
    {qual}""".format(
        all_cols=", ".join(all_cols),
        base_query=base_query,
        qual=qual,
    )


def powa_get_subscription_sample(subname=None):
    base_query = BASE_QUERY_SUBSCRIPTION_SAMPLE(subname)
    base_columns = ["srvid", "subname"]

    biggest = Biggest(base_columns, "ts")

    all_cols = base_columns + [
        "ts",
        biggest("ts", "'0 s'", "mesure_interval"),
        biggest("last_msg_lag"),
        biggest("report_lag"),
        biggest("apply_error_count"),
        biggest("sync_error_count"),
    ]

    return """SELECT {all_cols}
    FROM {base_query}""".format(
        all_cols=", ".join(all_cols),
        base_query=base_query,
    )


def powa_get_wal_sample():
    base_query = BASE_QUERY_WAL_SAMPLE
    base_columns = ["srvid"]

    biggest = Biggest(base_columns, "ts")
    biggestsum = Biggestsum(base_columns, "ts")

    all_cols = base_columns + [
        "ts",
        biggest("ts", "'0 s'", "mesure_interval"),
        biggestsum("wal_records"),
        biggestsum("wal_fpi"),
        biggestsum("wal_bytes"),
        biggestsum("wal_buffers_full"),
        biggestsum("wal_write"),
        biggestsum("wal_sync"),
        biggestsum("wal_write_time"),
        biggestsum("wal_sync_time"),
    ]

    return """SELECT {all_cols}
    FROM {base_query}
    GROUP BY {base_columns}, ts""".format(
        all_cols=", ".join(all_cols),
        base_columns=", ".join(base_columns),
        base_query=base_query,
    )


def powa_get_wal_receiver_sample():
    base_query = BASE_QUERY_WAL_RECEIVER_SAMPLE
    base_columns = ["srvid", "slot_name", "sender_host", "sender_port"]

    biggest = Biggest(base_columns, "ts")

    all_cols = base_columns + [
        "ts",
        biggest("ts", "'0 s'", "mesure_interval"),
        "slot_name",
        "sender_host",
        "sender_port",
        "pid",
        "status",
        "receive_start_lsn",
        "receive_start_tli",
        "last_received_lsn",
        biggest("last_received_lsn", label="received_bytes"),
        "written_lsn",
        "flushed_lsn",
        "received_tli",
        "last_msg_send_time",
        "last_msg_receipt_time",
        "latest_end_lsn",
        "latest_end_time",
        "conninfo",
    ]

    return """SELECT {all_cols}
    FROM {base_query}""".format(
        all_cols=", ".join(all_cols), base_query=base_query
    )


def powa_get_all_idx_sample(mode):
    if mode == "db":
        base_query = BASE_QUERY_ALL_IDXS_SAMPLE_DB
        base_columns = ["srvid", "dbid", "datname"]
    else:
        base_query = BASE_QUERY_ALL_IDXS_SAMPLE
        base_columns = ["srvid", "dbid", "datname", "relid"]

    biggest = Biggest(base_columns, "ts")
    biggestsum = Biggestsum(base_columns, "ts")

    all_cols = base_columns + [
        "ts",
        biggest("ts", "'0 s'", "mesure_interval"),
        "sum(idx_size) AS idx_size",
        biggestsum("idx_scan"),
        biggestsum("idx_tup_read"),
        biggestsum("idx_tup_fetch"),
        biggestsum("idx_blks_read"),
        biggestsum("idx_blks_hit"),
    ]

    return """SELECT {all_cols}
        FROM {base_query}
        WHERE srvid = %(server)s
        GROUP BY {base_columns}, ts""".format(
        all_cols=", ".join(all_cols),
        base_columns=", ".join(base_columns),
        base_query=base_query,
    )


def powa_get_all_tbl_sample(mode):
    if mode == "db":
        base_query = BASE_QUERY_ALL_TBLS_SAMPLE_DB
        base_columns = ["srvid", "dbid", "datname"]
    else:
        base_query = BASE_QUERY_ALL_TBLS_SAMPLE
        base_columns = ["srvid", "dbid", "datname", "relid"]

    biggest = Biggest(base_columns, "ts")
    biggestsum = Biggestsum(base_columns, "ts")

    all_cols = base_columns + [
        "ts",
        biggest("ts", "'0 s'", "mesure_interval"),
        "sum(tbl_size) AS tbl_size",
        biggestsum("seq_scan"),
        biggestsum("idx_scan"),
        biggestsum("n_tup_ins"),
        biggestsum("n_tup_upd"),
        biggestsum("n_tup_del"),
        biggestsum("n_tup_hot_upd"),
        biggestsum("vacuum_count"),
        biggestsum("autovacuum_count"),
        biggestsum("analyze_count"),
        biggestsum("autoanalyze_count"),
    ]

    return """SELECT {all_cols}
        FROM {base_query}
        WHERE srvid = %(server)s
        GROUP BY {base_columns}, ts""".format(
        all_cols=", ".join(all_cols),
        base_columns=", ".join(base_columns),
        base_query=base_query,
    )


def powa_get_user_fct_sample(mode):
    if mode == "db":
        base_query = BASE_QUERY_USER_FCTS_SAMPLE_DB
        base_columns = ["srvid", "dbid", "datname"]
    else:
        base_query = BASE_QUERY_USER_FCTS_SAMPLE
        base_columns = ["srvid", "dbid", "datname", "funcid"]

    biggest = Biggest(base_columns, "ts")
    biggestsum = Biggestsum(base_columns, "ts")

    all_cols = base_columns + [
        "ts",
        biggest("ts", "'0 s'", "mesure_interval"),
        biggestsum("calls"),
        biggestsum("total_time"),
        biggestsum("self_time"),
    ]

    return """SELECT {all_cols}
        FROM {base_query}
        WHERE srvid = %(server)s
        GROUP BY {base_columns}, ts""".format(
        all_cols=", ".join(all_cols),
        base_columns=", ".join(base_columns),
        base_query=base_query,
    )



class MenuEntry(object):
    def __init__(
        self,
        title,
        url_name,
        url_params=None,
        children_title=None,
        children=None,
    ):
        self.title = title
        self.url_name = url_name
        self.url_params = url_params or {}
        # Setting children helps showing dropdown menu in breadcrumb
        # Useful for listing databases for examples
        self.children_title = children_title
        self.children = children



"""
Global optimization widget
"""

from __future__ import absolute_import

import json
from powa.dashboards import MetricGroupDef, Widget
from powa.framework import AuthHandler
from powa.sql import (
    HypoIndex,
    get_any_sample_query,
    get_hypoplans,
    resolve_quals,
)
from powa.sql.views import qualstat_getstatdata
from psycopg2 import Error
from psycopg2.extras import RealDictCursor
from tornado.web import HTTPError


class IndexSuggestionHandler(AuthHandler):
    def post(self, srvid, database):
        try:
            # Check remote access first
            remote_conn = self.connect(
                srvid, database=database, remote_access=True
            )
            remote_cur = remote_conn.cursor()
        except Exception as e:
            raise HTTPError(
                501, "Could not connect to remote server: %s" % str(e)
            )

        payload = json.loads(self.request.body.decode("utf8"))
        from_date = payload["from_date"]
        to_date = payload["to_date"]
        indexes = []
        for ind in payload["indexes"]:
            hypoind = HypoIndex(ind["nspname"], ind["relname"], ind["ams"])
            hypoind._ddl = ind["ddl"]
            indexes.append(hypoind)
        queryids = payload["queryids"]
        powa_conn = self.connect(database="powa")
        cur = powa_conn.cursor(cursor_factory=RealDictCursor)
        cur.execute(
            """
            SELECT DISTINCT query, ps.queryid
            FROM {powa}.powa_statements ps
            WHERE srvid = %(srvid)s
            AND queryid IN %(queryids)s
        """,
            ({"srvid": srvid, "queryids": tuple(queryids)}),
        )
        queries = cur.fetchall()
        cur.close()
        # Create all possible indexes for this qual
        hypo_version = self.has_extension_version(
            srvid, "hypopg", "0.0.3", database=database
        )
        hypoplans = {}
        indbyname = {}
        inderrors = {}
        if hypo_version:
            # identify indexes
            # create them
            for ind in indexes:
                try:
                    remote_cur.execute(
                        """SELECT *
                            FROM hypopg_create_index(%(ddl)s)
                            """,
                        {"ddl": ind.ddl},
                    )
                    indname = remote_cur.fetchone()[1]
                    indbyname[indname] = ind
                except Error as e:
                    inderrors[ind.ddl] = str(e)
                    continue
                except Exception as e:
                    inderrors[ind.ddl] = str(e)
                    continue
            # Build the query and fetch the plans
            for row in queries:
                querystr = get_any_sample_query(
                    self, srvid, database, row["queryid"], from_date, to_date
                )
                if querystr:
                    try:
                        hypoplans[row["queryid"]] = get_hypoplans(
                            remote_cur, querystr, indbyname.values()
                        )
                    except Exception:
                        # TODO: stop ignoring the error
                        continue
            # To value of a link is the the reduction in cost
        remote_cur.close()
        result = {}
        result["plans"] = hypoplans
        result["inderrors"] = inderrors
        self.render_json(result)


class WizardMetricGroup(MetricGroupDef):
    """Metric group for the wizard."""

    name = "wizard"
    xaxis = "quals"
    axis_type = "category"
    data_url = r"/server/(\d+)/metrics/database/([^\/]+)/wizard/"
    enabled = False

    @property
    def query(self):
        pq = qualstat_getstatdata(eval_type="f")

        cols = [
            # queryid in pg11+ is int64, so the value can exceed javascript's
            # Number.MAX_SAFE_INTEGER, which means that the value can get
            # truncated by the browser, leading to looking for unexisting
            # queryid when processing this data.  To avoid that, simply cast
            # the value to text.
            "array_agg(cast(queryid AS text)) AS queryids",
            "qualid",
            "quals::jsonb AS quals",
            "occurences",
            "execution_count",
            "array_agg(query) AS queries",
            "avg_filter",
            "filter_ratio",
        ]

        query = """SELECT {cols}
        FROM (
            {pq}
        ) AS sub
        JOIN {{powa}}.powa_databases pd ON pd.oid = sub.dbid
            AND pd.srvid = sub.srvid
        WHERE pd.datname = %(database)s
        AND pd.srvid = %(server)s
        AND sub.avg_filter > 1000
        AND sub.filter_ratio > 0.3
        GROUP BY sub.qualid, sub.execution_count, sub.occurences,
            sub.quals::jsonb, sub.avg_filter, sub.filter_ratio
        ORDER BY sub.occurences DESC
        LIMIT 200""".format(
            cols=", ".join(cols),
            pq=pq,
        )
        return query

    def post_process(self, data, server, database, **kwargs):
        conn = self.connect(server, database=database, remote_access=True)
        data["data"] = resolve_quals(conn, data["data"])
        return data


class Wizard(Widget):
    def __init__(self, title):
        self.title = title

    def parameterized_json(self, handler, **parms):
        values = self.__dict__.copy()
        values["metrics"] = []
        values["type"] = "wizard"
        values["datasource"] = "wizard"

        # First check that we can connect on the remote server, otherwise we
        # won't be able to do anything
        try:
            handler.connect(
                parms["server"], database=parms["database"], remote_access=True
            )
        except Exception as e:
            values["has_remote_conn"] = False
            values["conn_error"] = str(e)
            return values

        values["has_remote_conn"] = True

        hypover = handler.has_extension_version(
            parms["server"], "hypopg", "0.0.3", database=parms["database"]
        )
        qsver = handler.has_extension_version(
            parms["server"], "pg_qualstats", "0.0.7"
        )
        values["has_hypopg"] = hypover
        values["has_qualstats"] = qsver
        values["server"] = parms["server"]
        values["database"] = parms["database"]
        return values



"""
Set of helper functions available from the templates.
"""

import json
import os
from datetime import datetime
from powa import __VERSION__
from powa.json import JSONEncoder
from tornado.options import options
from tornado.web import HTTPError


def version(_):
    """
    Returns:
        the current powa version.
    """
    return __VERSION__


def year(_):
    """
    Returns:
        the current year.
    """
    return datetime.now().year


def servers(_):
    """
    Returns:
        the servers defined in the configuration file.
    """
    return options.servers


def field(_, **kwargs):
    """
    Returns:
        a form field formatted with the given attributes.
    """
    kwargs.setdefault("tag", "input")
    kwargs.setdefault("type", "text")
    kwargs.setdefault("class", "form-control")
    attrs = " ".join(
        '%s="%s"' % (key, value)
        for key, value in kwargs.items()
        if key not in ("tag", "label")
    )
    kwargs["attrs"] = attrs

    def render(content):
        """
        Render the field itself.
        """
        kwargs["content"] = content.decode("utf8")
        return (
            """
<v-col cols="12">
    <label>%(label)s:
    <%(tag)s %(attrs)s>
        %(content)s
    </%(tag)s>
    </label>
</v-col>
"""
            % kwargs
        )

    return render


def flash(self, message, category=""):
    """
    Stores a message to be displayed on the next rendered page.
    """
    flashes = self.get_pickle_cookie("_flashes") or {}
    flashes.setdefault(category, []).append(message)
    self.set_pickle_cookie("_flashes", flashes)
    self.flashed_messages = flashes


def flashed_messages(self):
    """
    Returns:
        a mapping of flashed message category to their messages
    """
    messages = self.get_pickle_cookie("_flashes") or {}
    self.set_pickle_cookie("_flashes", None)
    for key, my_messages in self.flashed_messages.items():
        messages.setdefault(key, []).extend(my_messages)
    self.flashed_messages = {}
    return messages


def sanitycheck_messages(self):
    messages = {"error": []}

    # Check if now collector is running
    sql = """SELECT
        CASE WHEN val LIKE 'PoWA collector - main thread (%%'
            THEN 'Remote collector'
            ELSE 'Backgound worker'
        END AS powa_kind,
        date_trunc('second', backend_start) as start,
        datname, usename,
        coalesce(host(client_addr), '<local>') AS client_addr,
        count(datname) OVER () AS nb_found
        FROM (
            SELECT 'PoWA - %%' AS val
            UNION ALL
            SELECT 'PoWA collector - main thread (%%'
        ) n
        JOIN pg_stat_activity a ON a.application_name LIKE n.val"""
    rows = self.execute(sql)

    if rows is None:
        messages["error"].append("No collector is running!")

    sql = """SELECT
       CASE WHEN id = 0 THEN
          '<local>'
       ELSE
          COALESCE(alias, hostname || ':' || port)
       END AS alias,
        error
        FROM {powa}.powa_servers s
        JOIN (SELECT srvid, unnest(errors) error
            FROM {powa}.powa_snapshot_metas
            WHERE errors IS NOT NULL
        ) m ON m.srvid = s.id"""
    rows = self.execute(sql)

    if rows is not None and len(rows) > 0:
        for r in rows:
            messages["error"].append("%s: %s" % (r["alias"], r["error"]))
        return messages

    return {}


def to_json(_, value):
    """
    Utility function to render json in templates.
    """
    return JSONEncoder().encode(value)


def manifest(self, entrypoint):
    fn = os.path.realpath(__file__ + "../../static/dist/.vite/manifest.json")
    try:
        f = open(fn)
    except FileNotFoundError:
        raise HTTPError(
            500, "manifest.json doesn't exist, did you run `npm run build`?"
        )
    else:
        with f:
            entrypoints = json.load(f)
    return entrypoints[entrypoint]



"""
Dashboards for the various IO pages.
"""

from powa.config import ConfigChangesGlobal
from powa.dashboards import Dashboard, DashboardPage, Graph, Grid
from powa.io_template import TemplateIoGraph, TemplateIoGrid
from powa.server import ServerOverview


class TemplateIoOverview(DashboardPage):
    """
    Template dashboard for IO.
    """

    parent = ServerOverview
    timeline = ConfigChangesGlobal
    timeline_params = ["server"]

    def dashboard(self):
        # This COULD be initialized in the constructor, but tornado < 3 doesn't
        # call it
        if getattr(self, "_dashboard", None) is not None:
            return self._dashboard

        io_metrics = self.ds_graph.split(
            self,
            [
                ["reads", "writes", "writebacks", "extends", "fsyncs"],
                ["hits", "evictions", "reuses"],
            ],
        )
        graphs = []
        graphs.append(
            [
                Graph(
                    "IO blocks",
                    metrics=io_metrics[1],
                ),
                Graph(
                    "IO timing",
                    metrics=io_metrics[0],
                ),
                Graph(
                    "IO misc",
                    metrics=io_metrics[2],
                ),
            ]
        )

        graphs.append(
            [
                Grid(
                    "IO summary",
                    columns=[
                        {
                            "name": "backend_type",
                            "label": "Backend Type",
                            "url_attr": "backend_type_url",
                        },
                        {
                            "name": "obj",
                            "label": "Object Type",
                            "url_attr": "obj_url",
                        },
                        {
                            "name": "context",
                            "label": "Context",
                            "url_attr": "context_url",
                        },
                    ],
                    metrics=self.ds_grid.all(),
                )
            ]
        )

        self._dashboard = Dashboard(self.title, graphs)
        return self._dashboard


class ByBackendTypeIoGraph(TemplateIoGraph):
    """
    Metric group used by per backend_type graph.
    """

    name = "backend_type_io_graph"
    data_url = r"/server/(\d+)/metrics/backend_type_graph/([a-z0-9%]+)/io/"

    query_qual = "backend_type = %(backend_type)s"


class ByBackendTypeIoGrid(TemplateIoGrid):
    """
    Metric group used by per backend_type grid.
    """

    xaxis = "backend_type"
    name = "backend_type_io_grid"
    data_url = r"/server/(\d+)/metrics/backend_type_grid/([a-z0-9%]+)/io/"

    query_qual = "backend_type = %(backend_type)s"


class ByBackendTypeIoOverview(TemplateIoOverview):
    """
    Per backend-type Dashboard page.
    """

    base_url = r"/server/(\d+)/metrics/backend_type/([a-z0-9%]+)/io/overview/"
    params = ["server", "backend_type"]
    title = 'IO for "%(backend_type)s" backend type'

    ds_graph = ByBackendTypeIoGraph
    ds_grid = ByBackendTypeIoGrid
    datasources = [ds_graph, ds_grid]


class ByObjIoGraph(TemplateIoGraph):
    """
    Metric group used by per object graph.
    """

    name = "obj_io_graph"
    data_url = r"/server/(\d+)/metrics/obj_graph/([a-z0-9%]+)/io/"

    query_qual = "object = %(obj)s"


class ByObjIoGrid(TemplateIoGrid):
    """
    Metric group used by per object grid.
    """

    xaxis = "obj"
    name = "obj_io_grid"
    data_url = r"/server/(\d+)/metrics/obj_grid/([a-z0-9%]+)/io/"

    query_qual = "object = %(obj)s"


class ByObjIoOverview(TemplateIoOverview):
    """
    Per object Dashboard page.
    """

    base_url = r"/server/(\d+)/metrics/obj/([a-z0-9%]+)/io/overview/"
    params = ["server", "obj"]
    title = 'IO for "%(obj)s" object'

    ds_graph = ByObjIoGraph
    ds_grid = ByObjIoGrid
    datasources = [ds_graph, ds_grid]


class ByContextIoGraph(TemplateIoGraph):
    """
    Metric group used by per context graph.
    """

    name = "context_io_graph"
    data_url = r"/server/(\d+)/metrics/context_graph/([a-z0-9%]+)/io/"

    query_qual = "context = %(context)s"


class ByContextIoGrid(TemplateIoGrid):
    """
    Metric group used by per context grid.
    """

    xaxis = "context"
    name = "context_io_grid"
    data_url = r"/server/(\d+)/metrics/context_grid/([a-z0-9%]+)/io/"

    query_qual = "context = %(context)s"


class ByContextIoOverview(TemplateIoOverview):
    """
    Per context Dashboard page.
    """

    base_url = r"/server/(\d+)/metrics/context/([a-z0-9%]+)/io/overview/"
    params = ["server", "context"]
    title = 'IO for "%(context)s" context'

    ds_graph = ByContextIoGraph
    ds_grid = ByContextIoGrid
    datasources = [ds_graph, ds_grid]



"""
Module containing the by-database dashboard.
"""

from powa.config import ConfigChangesDatabase
from powa.dashboards import (
    ContentWidget,
    Dashboard,
    DashboardPage,
    Graph,
    Grid,
    MetricDef,
    MetricGroupDef,
    TabContainer,
)
from powa.framework import AuthHandler
from powa.server import ServerOverview
from powa.sql.utils import (
    block_size,
    mulblock,
    sum_per_sec,
    to_epoch,
    total_hit,
    total_read,
    wps,
)
from powa.sql.views_graph import (
    kcache_getstatdata_sample,
    powa_get_all_idx_sample,
    powa_get_all_tbl_sample,
    powa_get_database_sample,
    powa_get_pgsa_sample,
    powa_get_user_fct_sample,
    powa_getstatdata_sample,
    powa_getwaitdata_sample,
)
from powa.sql.views_grid import (
    powa_getstatdata_detailed_db,
    powa_getuserfuncdata_detailed_db,
    powa_getwaitdata_detailed_db,
)
from powa.wizard import Wizard, WizardMetricGroup
from tornado.web import HTTPError


class DatabaseSelector(AuthHandler):
    """Page allowing to choose a database."""

    def get(self):
        self.redirect(
            self.reverse_url(
                "DatabaseOverview",
                self.get_argument("server"),
                self.get_argument("database"),
            )
        )


class DatabaseOverviewMetricGroup(MetricGroupDef):
    """Metric group for the database global graphs."""

    name = "database_overview"
    xaxis = "ts"
    data_url = r"/server/(\d+)/metrics/database_overview/([^\/]+)/"
    avg_runtime = MetricDef(
        label="Avg runtime", type="duration", desc="Average query duration"
    )
    calls = MetricDef(
        label="Queries per sec",
        type="number",
        desc="Number of time the query has been executed, per second",
    )
    planload = MetricDef(
        label="Plantime per sec",
        type="duration",
        desc="Total planning duration",
    )
    load = MetricDef(
        label="Runtime per sec",
        type="duration",
        desc="Total duration of queries executed, per second",
    )
    total_blks_hit = MetricDef(
        label="Total shared buffers hit",
        type="sizerate",
        desc="Amount of data found in shared buffers",
    )
    total_blks_read = MetricDef(
        label="Total shared buffers miss",
        type="sizerate",
        desc="Amount of data found in OS cache or read from disk",
    )
    wal_records = MetricDef(
        label="#Wal records",
        type="integer",
        desc="Number of WAL records generated",
    )
    wal_fpi = MetricDef(
        label="#Wal FPI",
        type="integer",
        desc="Number of WAL full-page images generated",
    )
    wal_bytes = MetricDef(
        label="Wal bytes", type="size", desc="Amount of WAL bytes generated"
    )

    total_sys_hit = MetricDef(
        label="Total system cache hit",
        type="sizerate",
        desc="Amount of data found in OS cache",
    )
    total_disk_read = MetricDef(
        label="Total disk read",
        type="sizerate",
        desc="Amount of data read from disk",
    )
    minflts = MetricDef(
        label="Soft page faults",
        type="number",
        desc="Memory pages not found in the processor's MMU",
    )
    majflts = MetricDef(
        label="Hard page faults",
        type="number",
        desc="Memory pages not found in memory and loaded from storage",
    )
    # not maintained on GNU/Linux, and not available on Windows
    # nswaps = MetricDef(label="Swaps", type="number")
    # msgsnds = MetricDef(label="IPC messages sent", type="number")
    # msgrcvs = MetricDef(label="IPC messages received", type="number")
    # nsignals = MetricDef(label="Signals received", type="number")
    nvcsws = MetricDef(
        label="Voluntary context switches",
        type="number",
        desc="Number of voluntary context switches",
    )
    nivcsws = MetricDef(
        label="Involuntary context switches",
        type="number",
        desc="Number of involuntary context switches",
    )
    jit_functions = MetricDef(
        label="# of JIT functions",
        type="integer",
        desc="Total number of emitted functions",
    )
    jit_generation_time = MetricDef(
        label="JIT generation time",
        type="duration",
        desc="Total time spent generating code",
    )
    jit_inlining_count = MetricDef(
        label="# of JIT inlining",
        type="integer",
        desc="Number of queries where inlining was done",
    )
    jit_inlining_time = MetricDef(
        label="JIT inlining time",
        type="duration",
        desc="Total time spent inlining code",
    )
    jit_optimization_count = MetricDef(
        label="# of JIT optimization",
        type="integer",
        desc="Number of queries where optimization was done",
    )
    jit_optimization_time = MetricDef(
        label="JIT optimization time",
        type="duration",
        desc="Total time spent optimizing code",
    )
    jit_emission_count = MetricDef(
        label="# of JIT emission",
        type="integer",
        desc="Number of queries where emission was done",
    )
    jit_emission_time = MetricDef(
        label="JIT emission time",
        type="duration",
        desc="Total time spent emitting code",
    )
    jit_deform_count = MetricDef(
        label="# of JIT tuple deforming",
        type="integer",
        desc="Number of queries where tuple deforming was done",
    )
    jit_deform_time = MetricDef(
        label="JIT tuple deforming time",
        type="duration",
        desc="Total time spent deforming tuple",
    )
    jit_expr_time = MetricDef(
        label="JIT expression generation time",
        type="duration",
        desc="Total time spent generating expressions",
    )

    @classmethod
    def _get_metrics(cls, handler, **params):
        base = cls.metrics.copy()
        if not handler.has_extension(params["server"], "pg_stat_kcache"):
            for key in (
                "total_sys_hit",
                "total_disk_read",
                "minflts",
                "majflts",
                # "nswaps", "msgsnds", "msgrcvs", "nsignals",
                "nvcsws",
                "nivcsws",
            ):
                base.pop(key)
        else:
            base.pop("total_blks_read")

        if not handler.has_extension_version(
            params["server"], "pg_stat_statements", "1.8"
        ):
            for key in ("planload", "wal_records", "wal_fpi", "wal_bytes"):
                base.pop(key)

        if not handler.has_extension_version(
            handler.path_args[0], "pg_stat_statements", "1.10"
        ):
            for key in (
                "jit_functions",
                "jit_generation_time",
                "jit_inlining_count",
                "jit_inlining_time",
                "jit_optimization_count",
                "jit_optimization_time",
                "jit_emission_count",
                "jit_emission_time",
            ):
                base.pop(key)

        if not handler.has_extension_version(
            handler.path_args[0], "pg_stat_statements", "1.11"
        ):
            for key in (
                "jit_deform_count",
                "jit_deform_time",
                "jit_expr_time",
            ):
                base.pop(key)

        return base

    @property
    def query(self):
        # Fetch the base query for sample, and filter them on the database
        query = powa_getstatdata_sample("db", ["datname = %(database)s"])

        cols = [
            "srvid",
            to_epoch("ts"),
            sum_per_sec("calls", prefix="sub"),
            "sum(runtime) / greatest(sum(calls), 1) AS avg_runtime",
            sum_per_sec("runtime", prefix="sub", alias="load"),
            total_read("sub"),
            total_hit("sub"),
        ]

        if self.has_extension_version(
            self.path_args[0], "pg_stat_statements", "1.8"
        ):
            cols.extend(
                [
                    sum_per_sec("plantime", prefix="sub", alias="planload"),
                    sum_per_sec("wal_records", prefix="sub"),
                    sum_per_sec("wal_fpi", prefix="sub"),
                    sum_per_sec("wal_bytes", prefix="sub"),
                ]
            )

        if self.has_extension_version(
            self.path_args[0], "pg_stat_statements", "1.10"
        ):
            cols.extend(
                [
                    sum_per_sec("jit_functions"),
                    sum_per_sec("jit_generation_time"),
                    sum_per_sec("jit_inlining_count"),
                    sum_per_sec("jit_inlining_time"),
                    sum_per_sec("jit_optimization_count"),
                    sum_per_sec("jit_optimization_time"),
                    sum_per_sec("jit_emission_count"),
                    sum_per_sec("jit_emission_time"),
                ]
            )

        if self.has_extension_version(
            self.path_args[0], "pg_stat_statements", "1.11"
        ):
            cols.extend(
                [
                    sum_per_sec("jit_deform_count"),
                    sum_per_sec("jit_deform_time"),
                    sum_per_sec(
                        "jit_generation_time - jit_deform_time",
                        alias="jit_expr_time",
                    ),
                ]
            )

        from_clause = query

        if self.has_extension(self.path_args[0], "pg_stat_kcache"):
            from_clause = "({query}) AS sub2".format(query=query)

            # Add system metrics from pg_stat_kcache,
            kcache_query = kcache_getstatdata_sample(
                "db", ["datname = %(database)s"]
            )

            total_sys_hit = (
                "{total_read} - sum(sub.reads)"
                "/ greatest(extract(epoch FROM sub.mesure_interval), 1)"
                " AS total_sys_hit"
                "".format(total_read=total_read("sub", True))
            )
            total_disk_read = (
                "sum(sub.reads)"
                " / greatest(extract(epoch FROM sub.mesure_interval), 1)"
                " AS total_disk_read"
            )
            minflts = sum_per_sec("minflts", prefix="sub")
            majflts = sum_per_sec("majflts", prefix="sub")
            # nswaps = sum_per_sec('nswaps', prefix="sub")
            # msgsnds = sum_per_sec('msgsnds', prefix="sub")
            # msgrcvs = sum_per_sec('msgrcvs', prefix="sub")
            # nsignals = sum_per_sec(.nsignals', prefix="sub")
            nvcsws = sum_per_sec("nvcsws", prefix="sub")
            nivcsws = sum_per_sec("nivcsws", prefix="sub")

            cols.extend(
                [
                    total_sys_hit,
                    total_disk_read,
                    minflts,
                    majflts,
                    # nswaps, msgsnds, msgrcvs, nsignals,
                    nvcsws,
                    nivcsws,
                ]
            )

            from_clause += """
            LEFT JOIN ({kcache_query}) AS kc USING (dbid, ts, srvid)""".format(
                kcache_query=kcache_query
            )

        return """SELECT {cols}
        FROM (
            {from_clause}
        ) AS sub
        CROSS JOIN {bs}
        WHERE sub.calls != 0
        GROUP BY sub.srvid, sub.ts, block_size, sub.mesure_interval
        ORDER BY sub.ts""".format(
            cols=", ".join(cols), from_clause=from_clause, bs=block_size
        )


class DatabasePGSAOverview(MetricGroupDef):
    """
    Metric group used by pg_stat_activity graphs
    """

    name = "pgsa"
    xaxis = "ts"
    desc = "All backends, except autovacuum workers, are included"
    data_url = r"/server/(\d+)/metrics/pgsa_overview/([^\/]+)/"
    backend_xid_age = MetricDef(label="Backend xid age")
    backend_xmin_age = MetricDef(label="Backend xmin age")
    oldest_backend = MetricDef(label="Oldest backend", type="duration")
    oldest_xact = MetricDef(label="Oldest transaction", type="duration")
    oldest_query = MetricDef(label="Oldest query", type="duration")
    nb_idle = MetricDef(label="# of idle connections")
    nb_active = MetricDef(label="# of active connections")
    nb_idle_xact = MetricDef(label="# of idle in transaction connections")
    nb_fastpath = MetricDef(label="# of connections in fastpath function call")
    nb_idle_xact_abort = MetricDef(
        label="# of idle in transaction (aborted) connections"
    )
    nb_disabled = MetricDef(label="# of disabled connections")
    nb_unknown = MetricDef(label="# of connections in unknown state")
    nb_parallel_query = MetricDef(label="# of parallel queries")
    nb_parallel_worker = MetricDef(label="# of parallel workers")

    @classmethod
    def _get_metrics(cls, handler, **params):
        base = cls.metrics.copy()

        remote_pg_ver = handler.get_pg_version_num(handler.path_args[0])
        if remote_pg_ver is not None and remote_pg_ver < 130000:
            for key in ("nb_parallel_query", "nb_parallel_worker"):
                base.pop(key)
        return base

    @property
    def query(self):
        query = powa_get_pgsa_sample(per_db=True)
        # Those greatests are to avoid having negative values returned, see https://github.com/powa-team/powa-archivist/issues/98
        cols = [
            "extract(epoch FROM ts) AS ts",
            "greatest(max(backend_xid_age),0) AS backend_xid_age",
            "greatest(max(backend_xmin_age),0) AS backend_xmin_age",
            "greatest(max(backend_start_age) FILTER (WHERE datid IS NOT NULL),0) AS oldest_backend",
            "greatest(max(xact_start_age) FILTER (WHERE datid IS NOT NULL),0) AS oldest_xact",
            "greatest(max(query_start_age) FILTER (WHERE datid IS NOT NULL AND state !~ 'idle'),0) AS oldest_query",
            "count(*) FILTER (WHERE state = 'idle') AS nb_idle",
            "count(*) FILTER (WHERE state = 'active') AS nb_active",
            "count(*) FILTER (WHERE state = 'idle in transaction') AS nb_idle_xact",
            "count(*) FILTER (WHERE state = 'fastpath function call') AS nb_fastpath",
            "count(*) FILTER (WHERE state = 'idle in transaction (aborted)') AS nb_idle_xact_abort",
            "count(*) FILTER (WHERE state = 'disabled') AS nb_disabled",
            "count(*) FILTER (WHERE state IS NULL) AS nb_unknown",
            "count(DISTINCT leader_pid) AS nb_parallel_query",
            "count(*) FILTER (WHERE leader_pid IS NOT NULL) AS nb_parallel_worker",
        ]

        return """SELECT {cols}
            FROM ({query}) AS sub
            GROUP BY ts
            """.format(cols=", ".join(cols), query=query)


class DatabaseWaitOverviewMetricGroup(MetricGroupDef):
    """Metric group for the database global wait events graphs."""

    name = "database_waits_overview"
    xaxis = "ts"
    data_url = r"/server/(\d+)/metrics/database_waits_overview/([^\/]+)/"
    # pg 9.6 only metrics
    count_lwlocknamed = MetricDef(
        label="Lightweight Named",
        desc="Number of named lightweight lock wait events",
    )
    count_lwlocktranche = MetricDef(
        label="Lightweight Tranche",
        desc="Number of lightweight lock tranche wait events",
    )
    # pg 10+ metrics
    count_lwlock = MetricDef(
        label="Lightweight Lock",
        desc="Number of wait events due to lightweight locks",
    )
    count_lock = MetricDef(
        label="Lock", desc="Number of wait events due to heavyweight locks"
    )
    count_bufferpin = MetricDef(
        label="Buffer pin", desc="Number of wait events due to buffer pin"
    )
    count_activity = MetricDef(
        label="Activity",
        desc="Number of wait events due to postgres"
        " internal processes activity",
    )
    count_client = MetricDef(
        label="Client", desc="Number of wait events due to client activity"
    )
    count_extension = MetricDef(
        label="Extension",
        desc="Number wait events due to third-party extensions",
    )
    count_ipc = MetricDef(
        label="IPC",
        desc="Number of wait events due to inter-process communication",
    )
    count_timeout = MetricDef(
        label="Timeout", desc="Number of wait events due to timeouts"
    )
    count_io = MetricDef(
        label="IO", desc="Number of wait events due to IO operations"
    )

    def prepare(self):
        if not self.has_extension(self.path_args[0], "pg_wait_sampling"):
            raise HTTPError(501, "pg_wait_sampling is not installed")

    @property
    def query(self):
        query = powa_getwaitdata_sample("db", ["datname = %(database)s"])

        cols = [to_epoch("ts")]

        pg_version_num = self.get_pg_version_num(self.path_args[0])
        # if we can't connect to the remote server, assume pg10 or above
        if pg_version_num is not None and pg_version_num < 100000:
            cols += [
                wps("count_lwlocknamed"),
                wps("count_lwlocktranche"),
                wps("count_lock"),
                wps("count_bufferpin"),
            ]
        else:
            cols += [
                wps("count_lwlock"),
                wps("count_lock"),
                wps("count_bufferpin"),
                wps("count_activity"),
                wps("count_client"),
                wps("count_extension"),
                wps("count_ipc"),
                wps("count_timeout"),
                wps("count_io"),
            ]

        from_clause = "({query}) AS sub".format(query=query)

        return """SELECT {cols}
        FROM {from_clause}
        -- WHERE sub.count IS NOT NULL
        GROUP BY sub.ts, sub.mesure_interval
        ORDER BY sub.ts""".format(
            cols=", ".join(cols), from_clause=from_clause
        )


class DatabaseAllRelMetricGroup(MetricGroupDef):
    """
    Metric group used by "Database objects" graphs.
    """

    name = "all_relations"
    xaxis = "ts"
    data_url = r"/server/(\d+)/metrics/database_all_relations/([^\/]+)/"
    idx_size = MetricDef(
        label="Indexes size", type="size", desc="Size of all indexes"
    )
    tbl_size = MetricDef(
        label="Tables size", type="size", desc="Size of all tables"
    )
    idx_ratio = MetricDef(
        label="Index scans ratio",
        type="percent",
        desc="Ratio of index scan / seq scan",
    )
    idx_ratio = MetricDef(
        label="Index scans ratio",
        type="percent",
        desc="Ratio of index scan / seq scan",
    )
    idx_scan = MetricDef(
        label="Index scans",
        type="number",
        desc="Number of index scan per second",
    )
    seq_scan = MetricDef(
        label="Sequential scans",
        type="number",
        desc="Number of sequential scan per second",
    )
    n_tup_ins = MetricDef(
        label="Tuples inserted",
        type="number",
        desc="Number of tuples inserted per second",
    )
    n_tup_upd = MetricDef(
        label="Tuples updated",
        type="number",
        desc="Number of tuples updated per second",
    )
    n_tup_hot_upd = MetricDef(
        label="Tuples HOT updated",
        type="number",
        desc="Number of tuples HOT updated per second",
    )
    n_tup_del = MetricDef(
        label="Tuples deleted",
        type="number",
        desc="Number of tuples deleted per second",
    )
    vacuum_count = MetricDef(
        label="# Vacuum", type="number", desc="Number of vacuum per second"
    )
    autovacuum_count = MetricDef(
        label="# Autovacuum",
        type="number",
        desc="Number of autovacuum per second",
    )
    analyze_count = MetricDef(
        label="# Analyze", type="number", desc="Number of analyze per second"
    )
    autoanalyze_count = MetricDef(
        label="# Autoanalyze",
        type="number",
        desc="Number of autoanalyze per second",
    )

    @property
    def query(self):
        query1 = powa_get_all_tbl_sample("db")
        query2 = powa_get_all_idx_sample("db")

        cols = [
            "sub.srvid",
            "extract(epoch FROM sub.ts) AS ts",
            "sum(tbl_size) AS tbl_size",
            "sum(idx_size) AS idx_size",
            "CASE WHEN sum(sub.idx_scan + sub.seq_scan) = 0"
            " THEN 0"
            " ELSE sum(sub.idx_scan) * 100"
            " / sum(sub.idx_scan + sub.seq_scan)"
            " END AS idx_ratio",
            sum_per_sec("idx_scan", prefix="sub"),
            sum_per_sec("seq_scan", prefix="sub"),
            sum_per_sec("n_tup_ins", prefix="sub"),
            sum_per_sec("n_tup_upd", prefix="sub"),
            sum_per_sec("n_tup_hot_upd", prefix="sub"),
            sum_per_sec("n_tup_del", prefix="sub"),
            sum_per_sec("vacuum_count", prefix="sub"),
            sum_per_sec("autovacuum_count", prefix="sub"),
            sum_per_sec("analyze_count", prefix="sub"),
            sum_per_sec("autoanalyze_count", prefix="sub"),
        ]

        return """SELECT {cols}
        FROM (
            ({query1}) q1
            JOIN (SELECT srvid, dbid, ts, idx_size FROM ({query2})s ) q2
                USING (srvid, dbid, ts)
        ) AS sub
        WHERE sub.mesure_interval != '0 s'
        AND datname = %(database)s
        GROUP BY sub.srvid, sub.ts, sub.mesure_interval
        ORDER BY sub.ts""".format(
            cols=", ".join(cols), query1=query1, query2=query2
        )


class DatabaseUserFuncMetricGroup(MetricGroupDef):
    """
    Metric group used by "pg_stat_user_functions" graph.
    """

    name = "user_functions"
    xaxis = "ts"
    data_url = r"/server/(\d+)/metrics/database_user_functions/([^\/]+)/"
    calls = MetricDef(
        label="# of calls",
        type="number",
        desc="Number of function calls per second",
    )
    total_load = MetricDef(
        label="Total time per sec",
        type="number",
        desc="Total execution time duration",
    )
    self_load = MetricDef(
        label="Self time per sec",
        type="number",
        desc="Self execution time duration",
    )

    @property
    def query(self):
        query = powa_get_user_fct_sample("db")

        from_clause = query

        cols = [
            "sub.srvid",
            "extract(epoch FROM sub.ts) AS ts",
            sum_per_sec("calls"),
            sum_per_sec("total_time", alias="total_load"),
            sum_per_sec("self_time", alias="self_load"),
        ]

        return """SELECT {cols}
        FROM (
            {from_clause}
        ) AS sub
        WHERE sub.mesure_interval != '0 s'
        AND datname = %(database)s
        GROUP BY sub.srvid, sub.ts, sub.mesure_interval
        ORDER BY sub.ts""".format(
            cols=", ".join(cols), from_clause=from_clause
        )


class DatabaseDbActivityMetricGroup(MetricGroupDef):
    """
    Metric group used by "Database Activity" graphs.
    """

    name = "db_activity"
    xaxis = "ts"
    data_url = r"/server/(\d+)/metrics/db_activity/([^\/]+)/"
    numbackends = MetricDef(
        label="# of connections", desc="Total number of connections"
    )
    xact_commit = MetricDef(
        label="# of commits", desc="Total number of commits per second"
    )
    xact_rollback = MetricDef(
        label="# of rollbacks", desc="Total number of rollbacks per second"
    )
    conflicts = MetricDef(
        label="# of conflicts", desc="Total number of conflicts"
    )
    deadlocks = MetricDef(
        label="# of deadlocks", desc="Total number of deadlocks"
    )
    checksum_failures = MetricDef(
        label="# of checkum_failures", desc="Total number of checkum_failures"
    )
    session_time = MetricDef(
        label="Session time",
        type="duration",
        desc="Total time spent by database sessions per second",
    )
    active_time = MetricDef(
        label="Active time",
        type="duration",
        desc="Total time spent executing SQL statements per second",
    )
    idle_in_transaction_time = MetricDef(
        label="idle in xact time",
        type="duration",
        desc="Total time spent idling while in a transaction per second",
    )
    sessions = MetricDef(
        label="# sessions",
        desc="Total number of sessions established per second",
    )
    sessions_abandoned = MetricDef(
        label="# sessions abandoned",
        desc="Number of database sessions that "
        "were terminated because connection to "
        "the client was lost per second",
    )
    sessions_fatal = MetricDef(
        label="# sessions fatal",
        desc="Number of database sessions that "
        "were terminated by fatal errors per "
        "second",
    )
    sessions_killed = MetricDef(
        label="# sessions killed per second",
        desc="Number of database sessions that "
        "were terminated by operator intervention "
        "per second",
    )

    @classmethod
    def _get_metrics(cls, handler, **params):
        base = cls.metrics.copy()

        pg_version_num = handler.get_pg_version_num(handler.path_args[0])
        # if we can't connect to the remote server, assume pg14 or above
        if pg_version_num is not None and pg_version_num < 140000:
            for key in (
                "session_time",
                "active_time",
                "idle_in_transaction_time",
                "sessions",
                "sessions_abandoned",
                "sessions_fatal",
                "sessions_killed",
            ):
                base.pop(key)
        return base

    @property
    def query(self):
        query = powa_get_database_sample(True)

        cols = [
            "sub.srvid",
            "extract(epoch FROM sub.ts) AS ts",
            "numbackends",
            wps("xact_commit", do_sum=False),
            wps("xact_rollback", do_sum=False),
            "conflicts",
            "deadlocks",
            "checksum_failures",
            wps("session_time", do_sum=False),
            wps("active_time", do_sum=False),
            wps("idle_in_transaction_time", do_sum=False),
            wps("sessions", do_sum=False),
            wps("sessions_abandoned", do_sum=False),
            wps("sessions_fatal", do_sum=False),
            wps("sessions_killed", do_sum=False),
        ]

        return """SELECT {cols}
        FROM ({query}) sub
        WHERE sub.mesure_interval != '0 s'
        ORDER BY sub.ts""".format(
            cols=", ".join(cols),
            query=query,
        )


class ByQueryMetricGroup(MetricGroupDef):
    """Metric group for indivual query stats (displayed on the grid)."""

    name = "all_queries"
    xaxis = "queryid"
    axis_type = "category"
    data_url = r"/server/(\d+)/metrics/database_all_queries/([^\/]+)/"
    plantime = MetricDef(label="Plantime", type="duration")
    calls = MetricDef(label="#", type="integer")
    runtime = MetricDef(label="Time", type="duration", direction="descending")
    avg_runtime = MetricDef(label="Avg time", type="duration")
    shared_blks_read = MetricDef(label="Read", type="size")
    shared_blks_hit = MetricDef(label="Hit", type="size")
    shared_blks_dirtied = MetricDef(label="Dirtied", type="size")
    shared_blks_written = MetricDef(label="Written", type="size")
    temp_blks_read = MetricDef(label="Read", type="size")
    temp_blks_written = MetricDef(label="Written", type="size")
    blks_read_time = MetricDef(label="Total Read", type="duration")
    blks_write_time = MetricDef(label="Total Write", type="duration")
    wal_records = MetricDef(label="#Wal records", type="integer")
    wal_fpi = MetricDef(label="#Wal FPI", type="integer")
    wal_bytes = MetricDef(label="Wal bytes", type="size")
    jit_functions = MetricDef(label="#Functions", type="integer")
    jit_time = MetricDef(label="Time", type="duration")

    @classmethod
    def _get_metrics(cls, handler, **params):
        base = cls.metrics.copy()

        if not handler.has_extension_version(
            handler.path_args[0], "pg_stat_statements", "1.8"
        ):
            for key in ("plantime", "wal_records", "wal_fpi", "wal_bytes"):
                base.pop(key)

        if not handler.has_extension_version(
            handler.path_args[0], "pg_stat_statements", "1.10"
        ):
            base.pop("jit_functions")
            base.pop("jit_time")

        return base

    # TODO: refactor with GlobalDatabasesMetricGroup
    @property
    def query(self):
        # Working from the statdata detailed_db base query
        inner_query = powa_getstatdata_detailed_db()

        # Multiply each measure by the size of one block.
        cols = [
            "sub.srvid",
            "sub.queryid",
            "ps.query",
            "sum(sub.calls) AS calls",
            "sum(sub.runtime) AS runtime",
            mulblock("shared_blks_read", fn="sum"),
            mulblock("shared_blks_hit", fn="sum"),
            mulblock("shared_blks_dirtied", fn="sum"),
            mulblock("shared_blks_written", fn="sum"),
            mulblock("temp_blks_read", fn="sum"),
            mulblock("temp_blks_written", fn="sum"),
            "sum(sub.runtime) / greatest(sum(sub.calls), 1) AS avg_runtime",
            "sum(sub.shared_blk_read_time "
            + "+ sub.local_blk_read_time "
            + "+ sub.temp_blk_read_time) AS blks_read_time",
            "sum(sub.shared_blk_write_time "
            + "+ sub.local_blk_write_time "
            + "+ sub.temp_blk_write_time) AS blks_write_time",
        ]

        if self.has_extension_version(
            self.path_args[0], "pg_stat_statements", "1.8"
        ):
            cols.extend(
                [
                    "sum(sub.plantime) AS plantime",
                    "sum(sub.wal_records) AS wal_records",
                    "sum(sub.wal_fpi) AS wal_fpi",
                    "sum(sub.wal_bytes) AS wal_bytes",
                ]
            )

        if self.has_extension_version(
            self.path_args[0], "pg_stat_statements", "1.10"
        ):
            cols.extend(
                [
                    "sum(jit_functions) AS jit_functions",
                    "sum(jit_generation_time + jit_inlining_time"
                    + " + jit_optimization_time + jit_emission_time)"
                    + " AS jit_time",
                ]
            )

        from_clause = """(
                {inner_query}
            ) AS sub
            JOIN {{powa}}.powa_statements AS ps
                USING (srvid, queryid, userid, dbid)
            CROSS JOIN {bs}""".format(inner_query=inner_query, bs=block_size)

        return """SELECT {cols}
                FROM {from_clause}
                whERE datname = %(database)s
                GROUP BY srvid, queryid, query, block_size
                ORDER BY sum(runtime) DESC""".format(
            cols=", ".join(cols), from_clause=from_clause
        )

    def process(self, val, database=None, **kwargs):
        val["url"] = self.reverse_url(
            "QueryOverview", val["srvid"], database, val["queryid"]
        )
        return val


class ByQueryWaitSamplingMetricGroup(MetricGroupDef):
    """
    Metric group for indivual query wait events stats (displayed on the grid).
    """

    name = "all_queries_waits"
    xaxis = "query"
    axis_type = "category"
    data_url = r"/server/(\d+)/metrics/database_all_queries_waits/([^\/]+)/"
    counts = MetricDef(
        label="# of events", type="integer", direction="descending"
    )

    @property
    def query(self):
        # Working from the waitdata detailed_db base query
        inner_query = powa_getwaitdata_detailed_db()

        cols = [
            "srvid",
            "queryid",
            "ps.query",
            "event_type",
            "event",
            "sum(count) AS counts",
        ]

        from_clause = """(
            {inner_query}
        ) AS sub
            JOIN {{powa}}.powa_statements AS ps
                USING (srvid, queryid, dbid)""".format(inner_query=inner_query)

        return """SELECT {cols}
            FROM {from_clause}
            WHERE datname = %(database)s
            GROUP BY srvid, queryid, query, event_type, event
            ORDER BY sum(count) DESC""".format(
            cols=", ".join(cols), from_clause=from_clause
        )

    def process(self, val, database=None, **kwargs):
        val["url"] = self.reverse_url(
            "QueryOverview", val["srvid"], database, val["queryid"]
        )
        return val


class ByFuncUserFuncMetricGroup(MetricGroupDef):
    """
    Metric group for indivual function stats (displayed on the grid).
    """

    name = "all_functions_stats"
    xaxis = "funcname"
    axis_type = "category"
    data_url = r"/server/(\d+)/metrics/database_all_functions_stats/([^\/]+)/"
    lanname = MetricDef(label="Language", type="string")
    calls = MetricDef(
        label="# of calls", type="integer", direction="descending"
    )
    total_time = MetricDef(
        label="Cumulated total execution time", type="duration"
    )
    self_time = MetricDef(
        label="Cumulated self execution time", type="duration"
    )

    @property
    def query(self):
        # Working from the waitdata detailed_db base query
        inner_query = powa_getuserfuncdata_detailed_db()

        return inner_query + " ORDER BY calls DESC"

    def process(self, val, database=None, **kwargs):
        val["url"] = self.reverse_url(
            "FunctionOverview", val["srvid"], database, val["funcid"]
        )
        return val


class WizardThisDatabase(ContentWidget):
    title = "Apply wizardry to this database"

    data_url = r"/server/(\d+)/database/([^\/]+)/wizardthisdatabase/"

    def get(self, database):
        self.render_json(dict(url=self.reverse_url("WizardPage", database)))


class DatabaseOverview(DashboardPage):
    """DatabaseOverview Dashboard."""

    base_url = r"/server/(\d+)/database/([^\/]+)/overview"
    datasources = [
        DatabaseOverviewMetricGroup,
        ByQueryMetricGroup,
        ByQueryWaitSamplingMetricGroup,
        WizardMetricGroup,
        DatabaseWaitOverviewMetricGroup,
        ConfigChangesDatabase,
        DatabaseAllRelMetricGroup,
        DatabaseUserFuncMetricGroup,
        ByFuncUserFuncMetricGroup,
        DatabasePGSAOverview,
        DatabaseDbActivityMetricGroup,
    ]
    params = ["server", "database"]
    parent = ServerOverview
    title = "%(database)s"
    timeline = ConfigChangesDatabase

    def dashboard(self):
        # This COULD be initialized in the constructor, but tornado < 3 doesn't
        # call it
        if getattr(self, "_dashboard", None) is not None:
            return self._dashboard

        pgss18 = self.has_extension_version(
            self.path_args[0], "pg_stat_statements", "1.8"
        )
        pgss110 = self.has_extension_version(
            self.path_args[0], "pg_stat_statements", "1.10"
        )
        pgss111 = self.has_extension_version(
            self.path_args[0], "pg_stat_statements", "1.11"
        )

        self._dashboard = Dashboard("Database overview for %(database)s")

        db_metrics = [
            DatabaseOverviewMetricGroup.avg_runtime,
            DatabaseOverviewMetricGroup.load,
            DatabaseOverviewMetricGroup.calls,
        ]
        if pgss18:
            db_metrics.extend([DatabaseOverviewMetricGroup.planload])
        block_graph = Graph(
            "Blocks (On database %(database)s)",
            metrics=[DatabaseOverviewMetricGroup.total_blks_hit],
            color_scheme=None,
        )

        db_graphs = [
            [
                Graph("Calls (On database %(database)s)", metrics=db_metrics),
                block_graph,
            ]
        ]

        if "nb_parallel_query" in DatabasePGSAOverview._get_metrics(self):
            parallel_metrics = ["nb_parallel_query", "nb_parallel_worker"]
        else:
            parallel_metrics = []

        pgsa_metrics = DatabasePGSAOverview.split(
            self,
            [
                [
                    "backend_xid_age",
                    "backend_xmin_age",
                    "oldest_backend",
                    "oldest_xact",
                    "oldest_query",
                ],
                parallel_metrics,
            ],
        )
        db_graphs.append(
            [
                Graph(
                    "Global activity (On database %(database)s)",
                    metrics=pgsa_metrics[0],
                    desc=DatabasePGSAOverview.desc,
                    renderer="bar",
                    stack=True,
                )
            ]
        )
        if len(pgsa_metrics[2]) > 0:
            db_graphs[1].append(
                Graph(
                    "Parallel query (On database %(database)s)",
                    metrics=pgsa_metrics[2],
                )
            )
        db_graphs[1].append(
            Graph(
                "Backend age (On database %(database)s)",
                metrics=pgsa_metrics[1],
                desc=DatabasePGSAOverview.desc,
            )
        )

        graphs_dash = [Dashboard("General Overview", db_graphs)]
        graphs = [TabContainer("All databases", graphs_dash)]

        if pgss18:
            # Add WALs graphs
            wals_graphs = [
                [
                    Graph(
                        "WAL activity",
                        metrics=[
                            DatabaseOverviewMetricGroup.wal_records,
                            DatabaseOverviewMetricGroup.wal_fpi,
                            DatabaseOverviewMetricGroup.wal_bytes,
                        ],
                    ),
                ]
            ]
            graphs_dash.append(Dashboard("WALs", wals_graphs))

        # Add JIT graphs
        if pgss110:
            jit_tim = [
                DatabaseOverviewMetricGroup.jit_inlining_time,
                DatabaseOverviewMetricGroup.jit_optimization_time,
                DatabaseOverviewMetricGroup.jit_emission_time,
            ]

            if pgss111:
                jit_tim.extend(
                    [
                        DatabaseOverviewMetricGroup.jit_deform_time,
                        DatabaseOverviewMetricGroup.jit_expr_time,
                    ]
                )
            else:
                jit_tim.append(DatabaseOverviewMetricGroup.jit_generation_time)

            jit_cnt = [
                DatabaseOverviewMetricGroup.jit_functions,
                DatabaseOverviewMetricGroup.jit_inlining_count,
                DatabaseOverviewMetricGroup.jit_optimization_count,
                DatabaseOverviewMetricGroup.jit_emission_count,
            ]

            if pgss111:
                jit_cnt.append(DatabaseOverviewMetricGroup.jit_deform_count)

            jit_graphs = [
                [Graph("JIT timing", metrics=jit_tim, stack=True)],
                [Graph("JIT scheduling", metrics=jit_cnt)],
            ]

            graphs_dash.append(Dashboard("JIT", jit_graphs))

        # Add pg_stat_database graphs
        global_db_graphs = [
            [
                Graph(
                    "Transactions per second",
                    metrics=[
                        DatabaseDbActivityMetricGroup.xact_commit,
                        DatabaseDbActivityMetricGroup.xact_rollback,
                    ],
                    renderer="bar",
                    stack=True,
                ),
                Graph(
                    "Conflicts & deadlocks",
                    metrics=[
                        DatabaseDbActivityMetricGroup.conflicts,
                        DatabaseDbActivityMetricGroup.deadlocks,
                    ],
                ),
            ]
        ]
        if "sessions" in DatabaseDbActivityMetricGroup._get_metrics(self):
            global_db_graphs.append(
                [
                    Graph(
                        "Cumulated time per second",
                        metrics=[
                            DatabaseDbActivityMetricGroup.session_time,
                            DatabaseDbActivityMetricGroup.active_time,
                            DatabaseDbActivityMetricGroup.idle_in_transaction_time,
                        ],
                    ),
                    Graph(
                        "Sessions per second",
                        metrics=[
                            DatabaseDbActivityMetricGroup.sessions,
                            DatabaseDbActivityMetricGroup.sessions_abandoned,
                            DatabaseDbActivityMetricGroup.sessions_fatal,
                            DatabaseDbActivityMetricGroup.sessions_killed,
                        ],
                    ),
                ]
            )
        graphs_dash.append(Dashboard("Database activity", global_db_graphs))

        # Add powa_stat_all_relations graphs
        all_rel_graphs = [
            [
                Graph(
                    "Access pattern",
                    metrics=[
                        DatabaseAllRelMetricGroup.seq_scan,
                        DatabaseAllRelMetricGroup.idx_scan,
                        DatabaseAllRelMetricGroup.idx_ratio,
                    ],
                ),
                Graph(
                    "DML activity",
                    metrics=[
                        DatabaseAllRelMetricGroup.n_tup_del,
                        DatabaseAllRelMetricGroup.n_tup_hot_upd,
                        DatabaseAllRelMetricGroup.n_tup_upd,
                        DatabaseAllRelMetricGroup.n_tup_ins,
                    ],
                ),
            ],
            [
                Graph(
                    "Vacuum activity",
                    metrics=[
                        DatabaseAllRelMetricGroup.autoanalyze_count,
                        DatabaseAllRelMetricGroup.analyze_count,
                        DatabaseAllRelMetricGroup.autovacuum_count,
                        DatabaseAllRelMetricGroup.vacuum_count,
                    ],
                ),
                Graph(
                    "Object size",
                    metrics=[
                        DatabaseAllRelMetricGroup.tbl_size,
                        DatabaseAllRelMetricGroup.idx_size,
                    ],
                    renderer="bar",
                    stack=True,
                    color_scheme=["#73c03a", "#65b9ac"],
                ),
            ],
        ]
        graphs_dash.append(Dashboard("Database Objects", all_rel_graphs))

        # Add powa_stat_user_functions graphs
        user_fct_graph = [
            Graph(
                "User functions activity",
                metrics=DatabaseUserFuncMetricGroup.all(self),
            )
        ]
        user_fct_grid = [
            Grid(
                "User functions activity",
                columns=[
                    {
                        "name": "func_name",
                        "label": "Function name",
                        "url_attr": "url",
                    }
                ],
                metrics=ByFuncUserFuncMetricGroup.all(self),
            )
        ]
        graphs_dash.append(
            Dashboard("User functions", [user_fct_graph, user_fct_grid])
        )

        if self.has_extension(self.path_args[0], "pg_stat_kcache"):
            block_graph.metrics.insert(
                0, DatabaseOverviewMetricGroup.total_sys_hit
            )
            block_graph.metrics.insert(
                0, DatabaseOverviewMetricGroup.total_disk_read
            )
            block_graph.color_scheme = ["#cb513a", "#65b9ac", "#73c03a"]

            sys_graphs = [
                Graph(
                    "System resources (events per sec)",
                    url=self.docs_stats_url + "pg_stat_kcache.html",
                    metrics=[
                        DatabaseOverviewMetricGroup.majflts,
                        DatabaseOverviewMetricGroup.minflts,
                        # DatabaseOverviewMetricGroup.nswaps,
                        # DatabaseOverviewMetricGroup.msgsnds,
                        # DatabaseOverviewMetricGroup.msgrcvs,
                        # DatabaseOverviewMetricGroup.nsignals,
                        DatabaseOverviewMetricGroup.nvcsws,
                        DatabaseOverviewMetricGroup.nivcsws,
                    ],
                )
            ]

            graphs_dash.append(Dashboard("System resources", [sys_graphs]))
        else:
            block_graph.metrics.insert(
                0, DatabaseOverviewMetricGroup.total_blks_read
            )
            block_graph.color_scheme = ["#cb513a", "#73c03a"]

        if self.has_extension(self.path_args[0], "pg_wait_sampling"):
            metrics = None
            pg_version_num = self.get_pg_version_num(self.path_args[0])
            # if we can't connect to the remote server, assume pg10 or above
            if pg_version_num is not None and pg_version_num < 100000:
                metrics = [
                    DatabaseWaitOverviewMetricGroup.count_lwlocknamed,
                    DatabaseWaitOverviewMetricGroup.count_lwlocktranche,
                    DatabaseWaitOverviewMetricGroup.count_lock,
                    DatabaseWaitOverviewMetricGroup.count_bufferpin,
                ]
            else:
                metrics = [
                    DatabaseWaitOverviewMetricGroup.count_lwlock,
                    DatabaseWaitOverviewMetricGroup.count_lock,
                    DatabaseWaitOverviewMetricGroup.count_bufferpin,
                    DatabaseWaitOverviewMetricGroup.count_activity,
                    DatabaseWaitOverviewMetricGroup.count_client,
                    DatabaseWaitOverviewMetricGroup.count_extension,
                    DatabaseWaitOverviewMetricGroup.count_ipc,
                    DatabaseWaitOverviewMetricGroup.count_timeout,
                    DatabaseWaitOverviewMetricGroup.count_io,
                ]

            graphs_dash.append(
                Dashboard(
                    "Wait Events",
                    [
                        [
                            Graph(
                                "Wait Events (per second)",
                                url=self.docs_stats_url
                                + "pg_wait_sampling.html",
                                metrics=metrics,
                            )
                        ]
                    ],
                )
            )

        toprow = [
            {
                # query
            }
        ]

        if pgss18:
            toprow.extend(
                [
                    {
                        # plan time
                    }
                ]
            )

        toprow.extend(
            [
                {"name": "Execution", "colspan": 3},
                {
                    "name": "Blocks",
                    "colspan": 4,
                },
                {"name": "Temp blocks", "colspan": 2},
                {"name": "I/O Time", "colspan": 2},
            ]
        )

        if pgss18:
            toprow.extend([{"name": "WALs", "colspan": 3}])

        if pgss110:
            toprow.extend([{"name": "JIT", "colspan": 2}])
        self._dashboard.widgets.extend(
            [
                graphs,
                [
                    Grid(
                        "Details for all queries",
                        toprow=toprow,
                        columns=[
                            {
                                "name": "query",
                                "label": "Query",
                                "type": "query",
                                "url_attr": "url",
                                "max_length": 70,
                            }
                        ],
                        metrics=ByQueryMetricGroup.all(self),
                    )
                ],
            ]
        )

        if self.has_extension(self.path_args[0], "pg_wait_sampling"):
            self._dashboard.widgets.extend(
                [
                    [
                        Grid(
                            "Wait events for all queries",
                            url=self.docs_stats_url + "pg_wait_sampling.html",
                            columns=[
                                {
                                    "name": "query",
                                    "label": "Query",
                                    "type": "query",
                                    "url_attr": "url",
                                    "max_length": 70,
                                },
                                {
                                    "name": "event_type",
                                    "label": "Event Type",
                                },
                                {
                                    "name": "event",
                                    "label": "Event",
                                },
                            ],
                            metrics=ByQueryWaitSamplingMetricGroup.all(),
                        )
                    ]
                ]
            )

        self._dashboard.widgets.extend([[Wizard("Index suggestions")]])
        return self._dashboard

    @classmethod
    def breadcrum_title(cls, handler, param):
        return param["database"]
